import logging
import time

import numpy as np

from kite.quadtree import QuadNode
from kite.scene import TestScene
from kite.scene_stack import SceneStack, TSScene
from kite.spool import spool

logging.basicConfig(level=logging.DEBUG)
QuadNode.MIN_PIXEL_LENGTH_NODE = 32

t0 = time.time()
dt = 60 * 60 * 24 * 365.25 / 2


class TSTestScene(TestScene, TSScene):
    @classmethod
    def createSine(cls, *args, **kwargs):
        return super().createSine(*args, **kwargs)

    @classmethod
    def createRandom(cls, *args, **kwargs):
        return super().createRandom(*args, **kwargs)

    @classmethod
    def createGauss(cls, *args, **kwargs):
        return super().createGauss(*args, **kwargs)

    @classmethod
    def createFractal(cls, *args, **kwargs):
        return super().createFractal(*args, **kwargs)


nscenes = 10
scenes = {}

for isc in range(nscenes):
    t = t0 - isc * dt
    sc = TSTestScene.createRandom(256, 256)
    sc.addNoise(1.0)

    scenes[t] = sc


stack = SceneStack()
for ts, scene in scenes.items():
    stack.add_scene(scene, timestamp=ts)



import glob
import os
import re
import time
from datetime import datetime

import numpy as np
import scipy.io
import utm

from kite import util

__all__ = ["Gamma", "Matlab", "ISCE", "GMTSAR", "ROI_PAC", "SARscape", "SNAP_Gamma"]

try:
    from osgeo import gdal

    __all__.append("LiCSAR")
    __all__.append("ARIA")
except ImportError:
    pass


d2r = np.pi / 180.0
km = 1e3
op = os.path

LAMBDA_SENTINEL = 0.055465763


def check_required(required, params):
    for r in required:
        if r not in params:
            return False
    return True


def safe_cast(val, to_type, default=None):
    try:
        return to_type(val)
    except (ValueError, TypeError):
        return default


class HeaderError(Exception):
    pass


class AttribDict(dict):
    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, item, value):
        self[item] = value


class SceneIO(object):
    """Prototype class for SARIO objects."""

    def __init__(self, scene=None):
        if scene is not None:
            self._log = scene._log.getChild("IO/%s" % self.__class__.__name__)
        else:
            import logging

            self._log = logging.getLogger("SceneIO/%s" % self.__class__.__name__)

        self.container = AttribDict(
            phi=0.0,  # Look orientation counter-clockwise angle from east
            theta=0.0,  # Look elevation angle (up from horizontal in degree)
            # 90 deg North
            displacement=None,  # Displacement towards LOS
            frame=AttribDict(
                llLon=None,  # Lower left corner latitude
                llLat=None,  # Lower left corner londgitude
                dN=None,  # Pixel delta in north, meter or degree
                dE=None,  # Pixel delta in east, meter or degree
                spacing="meter",  # Pixel spacing unit
            ),
            # Meta information
            meta=AttribDict(
                title=None,
                orbital_node=None,
                satellite_name=None,
                wavelength=None,
                time_master=None,
                time_slave=None,
            ),
            # All extra information
            extra={},
        )

    def read(self, filename, **kwargs):
        """Read function of the file format

        :param filename: file to read
        :type filename: string
        :param kwargs: Keyword arguments
        :type kwargs: {dict}
        """
        raise NotImplementedError("read not implemented")

    def write(self, filename, **kwargs):
        """Write method for IO

        :param filename: file to write to
        :type filename: string
        :param **kwargs: Keyword arguments
        :type **kwargs: {dict}
        """
        raise NotImplementedError("write not implemented")

    def validate(self, filename, **kwargs):
        """Validate file format

        :param filename: file to validate
        :type filename: string
        :returns: Validation
        :rtype: {bool}
        """
        raise NotImplementedError("validate not implemented")


class Matlab(SceneIO):
    """
    Variable naming conventions for Matlab :file:`.mat` container:

        ================== ==================== ===================== =====
        Property           Matlab ``.mat`` name type                  unit
        ================== ==================== ===================== =====
        Scene.displacement ``ig_``              n x m array           [m]
        Scene.phi          ``phi``              float or n x m array  [rad]
        Scene.theta        ``theta``            float or n x m array  [rad]
        Scene.frame.x      ``xx``               n x 1 vector          [m]
        Scene.frame.y      ``yy``               m x 1 vector          [m]
        Scene.utm_zone     ``utm_zone``         str ('33T')
        ================== ==================== ===================== =====

    Displacement is expected to be in meters. Note that the displacement maps
    could also be pixel offset maps rather than unwrapped SAR interferograms.
    For SAR azimuth pixel offset maps calculate ``phi`` from the heading
    direction and set ``theta=0.``. For SAR range pixel offsets use the same
    LOS angles as for InSAR.
    """

    def validate(self, filename, **kwargs):
        if filename[-4:] == ".mat":
            return True
        else:
            return False
        try:
            variables = self.io.whosmat(filename)
            if len(variables) > 50:
                return False
            return True
        except ValueError:
            return False

    def read(self, filename, **kwargs):
        container = self.container

        mat = scipy.io.loadmat(filename)
        utm_e = None
        utm_n = None
        utm_zone = None
        utm_zone_letter = None
        phi0 = None
        theta0 = None

        for mat_k, _ in mat.items():
            for io_k in container.keys():
                if io_k in mat_k:
                    container[io_k] = np.rot90(mat[mat_k])
                elif "ig_" in mat_k:
                    container.displacement = np.rot90(mat[mat_k])
                elif "xx" in mat_k:
                    utm_e = mat[mat_k].flatten()
                elif "yy" in mat_k:
                    utm_n = mat[mat_k].flatten()
                elif "utm_zone" in mat_k:
                    utm_zone = int(mat["utm_zone"][0][:-1])
                    utm_zone_letter = str(mat["utm_zone"][0][-1])
                elif "phi" in mat_k:
                    phi0 = mat[mat_k].flatten()
                elif "theta" in mat_k:
                    theta0 = mat[mat_k].flatten()

        if len(theta0) == 1:
            container.theta = np.ones(np.shape(container.displacement)) * theta0

        if len(theta0) == 1:
            container.phi = np.ones(np.shape(container.displacement)) * phi0

        if utm_zone is None:
            utm_zone = 33
            utm_zone_letter = "N"
            self._log.warning(
                "Variable utm_zone not defined. Defaulting to UTM Zone %d%s!"
                % (utm_zone, utm_zone_letter)
            )

        if not (np.all(utm_e) or np.all(utm_n)):
            self._log.warning("Could not find referencing UTM vectors in .mat file!")
            utm_e = np.linspace(100000, 110000, container.displacement.shape[0])
            utm_n = np.linspace(1100000, 1110000, container.displacement.shape[1])

        if utm_e.min() < 1e4 or utm_n.min() < 1e4:
            utm_e *= km
            utm_n *= km

        container.frame.dE = np.abs(utm_e[1] - utm_e[0])
        container.frame.dN = np.abs(utm_n[1] - utm_n[0])
        try:
            container.frame.llLat, container.frame.llLon = utm.to_latlon(
                utm_e.min(), utm_n.min(), utm_zone, utm_zone_letter
            )

        except utm.error.OutOfRangeError:
            self._log.warning(
                "Could not interpret spatial vectors," " referencing to 0, 0 (lat, lon)"
            )
            container.frame.llLat, container.frame.llLon = (0.0, 0.0)
        return container


class Gamma(SceneIO):
    """

    Reading geocoded displacement maps (unwrapped igs) originating
        from GAMMA software.

    .. note :: Expects:

        * [:file:`*`] Binary file from Gamma with displacement in radians
        * [:file:`*.slc.par`] If you want to translate radians to
          meters using the `radar_frequency`.
        * [:file:`*par`] Parameter file, describing ``corner_lat, corner_lon,
          nlines, width, post_lat, post_lon`` or ``post_north, post_east,
          corner_east, corner_north, nlines, width``.
        * [:file:`*theta*`, :file:`*phi*`] Two look vector files,
          generated by GAMMA command ``look_vector``.

    .. warning ::

        * Data has to be georeferenced to latitude/longitude or UTM!
        * Look vector files - expected to have a particular name
    """

    @staticmethod
    def _parseParameterFile(filename):
        params = {}
        rc = re.compile(r"^(\w*):\s*([a-zA-Z0-9+-.*]*\s[a-zA-Z0-9_]*).*")

        with open(filename, mode="r") as par:
            for line in par:
                parsed = rc.match(line)
                if parsed is None:
                    continue

                groups = parsed.groups()
                params[groups[0]] = safe_cast(
                    groups[1], float, default=groups[1].strip()
                )
        return params

    def _getParameters(self, path, log=False):
        required_utm = [
            "post_north",
            "post_east",
            "corner_east",
            "corner_north",
            "nlines",
            "width",
        ]
        required_lat_lon = [
            "corner_lat",
            "corner_lon",
            "nlines",
            "width",
            "post_lat",
            "post_lon",
        ]

        path = op.dirname(op.realpath(path))
        par_files = glob.glob("%s/*par" % path)

        for file in par_files:
            params = self._parseParameterFile(file)

            if check_required(required_utm, params) or check_required(
                required_lat_lon, params
            ):
                if not log:
                    self._log.info("Found parameter file %s", file)
                return params

        raise ImportError("Parameter file does not hold required parameters")

    def _getSLCParameters(self, path):
        required_params = ("nlines",)
        path = op.dirname(op.realpath(path))
        par_files = glob.glob(f"{path}/*par")

        for file in par_files:
            params = self._parseParameterFile(file)

            if check_required(required_params, params):
                self._log.info("Found SLC parameter file %s", file)
                return params

        raise ImportError(
            "Could not find SLC parameter file *.par"
            f" with parameters {required_params}"
        )

    def validate(self, filename, **kwargs):
        try:
            par_file = kwargs.pop("par_file", filename)
            self._getParameters(par_file)
            return True
        except ImportError:
            return False

    def _getLOSAngles(self, filename, pattern):
        path = op.dirname(op.realpath(filename))
        phi_files = glob.glob("%s/%s" % (path, pattern))
        if len(phi_files) == 0:
            self._log.warning(
                "Could not find LOS file %s, " "defaulting to angle to 0." % pattern
            )
            return 0.0
        elif len(phi_files) > 1:
            self._log.warning(
                "Found multiple LOS files %s, " "defaulting to angle 0." % pattern
            )
            return 0.0

        filename = phi_files[0]
        self._log.info("Loading LOS %s from %s" % (pattern, filename))
        return np.memmap(filename, mode="r", dtype=">f4")

    def read(self, filename, **kwargs):
        """
        :param filename: Gamma software parameter file
        :type filename: str
        :param par_file: Corresponding parameter (:file:`*par`) file.
                         (optional)
        :type par_file: str
        :returns: Import dictionary
        :rtype: dict
        :raises: ImportError
        """
        par_file = kwargs.pop("par_file", filename)

        params = self._getParameters(par_file, log=True)

        try:
            params_slc = self._getSLCParameters(par_file)
        except ImportError as exc:
            raise exc

        fill = None

        ncols = int(params["width"])
        nlines = int(params["nlines"])
        radar_frequency = float(params_slc.get("radar_frequency", 5.405e9))  # Sentinel1

        displ = np.fromfile(filename, dtype=">f4")
        # Resize array if last line is not scanned completely
        if (displ.size % ncols) != 0:
            fill = np.empty(ncols - displ.size % ncols)
            fill.fill(np.nan)
            displ = np.append(displ, fill)

        displ = displ.reshape(nlines, ncols)
        displ[displ == -0.0] = np.nan
        displ = np.flipud(displ)

        if radar_frequency is not None:
            radar_frequency = float(radar_frequency)
            self._log.info(
                "Scaling displacement by radar_frequency %f GHz"
                % (radar_frequency / 1e9)
            )
            wavelength = util.C / radar_frequency
            displ /= -4 * np.pi
            displ *= wavelength

        else:
            wavelength = "None"
            self._log.warning(
                "Could not determine radar_frequency from *.slc.par file!"
                " Leaving displacement to radians."
            )

        phi = self._getLOSAngles(filename, "*phi*")
        theta = self._getLOSAngles(filename, "*theta*")
        theta = theta

        if isinstance(phi, np.ndarray):
            phi = phi.reshape(nlines, ncols)
            phi = np.flipud(phi)
        if isinstance(theta, np.ndarray):
            theta = theta.reshape(nlines, ncols)
            theta = np.flipud(theta)

        if fill is not None:
            theta = np.append(theta, fill)
            phi = np.append(phi, fill)

        container = self.container

        container.displacement = displ
        container.theta = theta
        container.phi = phi

        container.meta.wavelength = wavelength
        container.meta.title = params.get("title", "None")

        container.bin_file = filename
        container.par_file = par_file

        if params["DEM_projection"] == "UTM":
            utm_zone = params["projection_zone"]
            try:
                utm_zone_letter = utm.latitude_to_zone_letter(params["center_latitude"])
            except ValueError:
                self._log.warning(
                    "Could not parse UTM Zone letter," " defaulting to N!"
                )
                utm_zone_letter = "N"

            self._log.info(
                "Using UTM reference: Zone %d%s" % (utm_zone, utm_zone_letter)
            )

            dN = params["post_north"]
            dE = params["post_east"]

            utm_corn_e = params["corner_east"]
            utm_corn_n = params["corner_north"]

            utm_corn_eo = utm_corn_e + dE * displ.shape[1]
            utm_corn_no = utm_corn_n + dN * displ.shape[0]

            utm_e = np.linspace(utm_corn_e, utm_corn_eo, displ.shape[1])
            utm_n = np.linspace(utm_corn_n, utm_corn_no, displ.shape[0])

            llLat, llLon = utm.to_latlon(
                utm_e.min(), utm_n.min(), utm_zone, utm_zone_letter
            )

            container.frame.llLat = llLat
            container.frame.llLon = llLon

            container.frame.dE = abs(dE)
            container.frame.dN = abs(dN)

        else:
            self._log.info("Using Lat/Lon reference")
            container.frame.spacing = "degree"
            container.frame.llLat = params["corner_lat"] + params["post_lat"] * nlines
            container.frame.llLon = params["corner_lon"]
            container.frame.dE = abs(params["post_lon"])
            container.frame.dN = abs(params["post_lat"])

        return container


class ROI_PAC(SceneIO):
    """
    .. note:: Expects:

        * Binary file from ROI_PAC (:file:`*`)
        * Parameter file (:file:`<binary_file>.rsc`),
          describing ``WIDTH, FILE_LENGTH, X_FIRST, Y_FIRST, X_STEP,
          Y_STEP, WAVELENGTH``
        * If the georeferencing is in UTM coordinates, further needed
          entries in parameter file are 'X_UNIT' and 'Y_UNIT' that give
          'meters' and 'LAT_REF3' as well as 'LON_REF3'.

        The unwrapped displacement is expected in radians and will be scaled
        to meters by ``WAVELENGTH`` parsed from the :file:`*.rsc` file.

    """

    def validate(self, filename, **kwargs):
        try:
            par_file = kwargs.pop("par_file", self._getParameterFile(filename))
            self._parseParameterFile(par_file)
            return True
        except ImportError:
            return False

    def _getParameterFile(self, bin_file):
        par_file = op.realpath(bin_file) + ".rsc"
        try:
            self._parseParameterFile(par_file)
            self._log.info("Found parameter file %s" % par_file)
            return par_file
        except (ImportError, IOError):
            raise ImportError("Could not find ROI_PAC parameter file (%s)" % par_file)

    @staticmethod
    def _parseParameterFile(par_file):
        params = {}
        required_L0 = [
            "WIDTH",
            "FILE_LENGTH",
            "X_FIRST",
            "Y_FIRST",
            "X_STEP",
            "Y_STEP",
            "WAVELENGTH",
        ]
        required_utm = ["X_UNIT", "LAT_REF1", "LON_REF1"]

        rc = re.compile(r"([\w]*)\s*([\w.+-]*)")
        with open(par_file, "r") as par:
            for line in par:
                parsed = rc.match(line)
                if parsed is None:
                    continue
                groups = parsed.groups()
                params[groups[0]] = safe_cast(
                    groups[1], float, default=groups[1].strip()
                )

        if check_required(required_L0, params):
            if check_required(required_utm, params):
                geo_ref = "all"
                return params, geo_ref
            else:
                geo_ref = "latlon"
                return params, geo_ref

        raise ImportError(
            "Parameter file %s does not hold the basic \
             required parameters"
            % par_file
        )

    def read(self, filename, **kwargs):
        """
        :param filename: ROI_PAC binary file
        :type filename: str
        :param par_file: Corresponding parameter (:file:`*rsc`) file.
                         (optional)
        :type par_file: str
        :returns: Import dictionary
        :rtype: dict
        :raises: ImportError
        """
        par_file = kwargs.pop("par_file", self._getParameterFile(filename))

        par, geo_ref = self._parseParameterFile(par_file)
        nlines = int(par["FILE_LENGTH"])
        ncols = int(par["WIDTH"])
        wavelength = par["WAVELENGTH"]
        heading = par["HEADING_DEG"]
        if geo_ref == "latlon":
            lat_ref = par["Y_FIRST"]
            lon_ref = par["X_FIRST"]
        elif geo_ref == "all":
            lat_ref = par["LAT_REF3"]
            lon_ref = par["LON_REF3"]

        look_ref1 = par["LOOK_REF1"]
        look_ref2 = par["LOOK_REF2"]
        look_ref3 = par["LOOK_REF3"]
        look_ref4 = par["LOOK_REF4"]

        utm_zone_letter = utm.latitude_to_zone_letter(lat_ref)
        utm_zone = utm.latlon_to_zone_number(lat_ref, lon_ref)

        look = np.mean(np.array([look_ref1, look_ref2, look_ref3, look_ref4]))

        data = np.memmap(filename, dtype="<f4")
        data = data.reshape(nlines, ncols * 2)

        displ = data[:, ncols:]
        displ = np.flipud(displ)
        displ[displ == -0.0] = np.nan
        displ = displ / (4.0 * np.pi) * wavelength

        z_scale = par.get("Z_SCALE", 1.0)
        z_offset = par.get("Z_OFFSET", 0.0)
        displ += z_offset
        displ *= z_scale

        container = self.container

        container.displacement = displ
        container.theta = np.deg2rad(90.0 - look)
        container.phi = np.deg2rad(-heading + 180.0)

        container.meta.title = par.get("TITLE", "None")
        container.meta.wavelength = par["WAVELENGTH"]
        container.bin_file = filename
        container.par_file = par_file

        if geo_ref == "all":
            if par["X_UNIT"] == "meters":
                container.frame.spacing = "meter"
                container.frame.dE = par["X_STEP"]
                container.frame.dN = -par["Y_STEP"]
                geo_ref = "utm"

            elif par["X_UNIT"] == "degree":
                container.frame.spacing = "degree"
                geo_ref = "latlon"

        elif geo_ref == "latlon":
            self._log.info("Georeferencing is in Lat-Lon [degrees].")
            container.frame.spacing = "degree"
            container.frame.llLat = par["Y_FIRST"] + par["Y_STEP"] * nlines
            container.frame.llLon = par["X_FIRST"]

            # c_utm_0 = utm.from_latlon(lat_ref, lon_ref)
            # c_utm_1 = utm.from_latlon(lat_ref + par['Y_STEP'],
            #                           lon_ref + par['X_STEP'])

            # c.frame.dE = c_utm_1[0] - c_utm_0[0]
            # c.frame.dN = abs(c_utm_1[1] - c_utm_0[1])
            container.frame.dE = par["X_STEP"]
            container.frame.dN = -par["Y_STEP"]

        elif geo_ref == "utm":
            self._log.info(
                "Georeferencing is in UTM (zone %d%s)", utm_zone, utm_zone_letter
            )
            y_ll = par["Y_FIRST"] + par["Y_STEP"] * nlines
            container.frame.llLat, container.frame.llLon = utm.to_latlon(
                par["X_FIRST"], y_ll, utm_zone, zone_letter=utm_zone_letter
            )

        return self.container


class ISCEXMLParser(object):
    def __init__(self, filename):
        import xml.etree.ElementTree as ET

        self.root = ET.parse(filename).getroot()

    @staticmethod
    def type_convert(value):
        for t in (float, int, str):
            try:
                return t(value)
            except ValueError:
                continue
        raise ValueError("Could not convert value")

    def getProperty(self, name):
        name = name.lower()

        for child in self.root.iter():
            child_name = child.get("name")
            if isinstance(child_name, str):
                child_name = child_name.lower()
            if child_name == name.lower():
                if child.tag == "property":
                    return self.type_convert(child.find("value").text)
                elif child.tag == "component":
                    values = {}
                    for prop in child.iter("property"):
                        values[prop.get("name")] = self.type_convert(
                            prop.find("value").text
                        )
                    return values
        return None


class ISCE(SceneIO):
    """
    Reading geocoded, unwrapped displacement maps
    processed with ISCE software (https://winsar.unavco.org/isce.html).

    .. note :: Expects:

        * Unwrapped displacement binary (:file:`*.unw.geo`)
        * Metadata XML (:file:`*.unw.geo.xml`)
        * LOS binary data (:file:`*.rdr.geo`)

    .. note ::

        When using ``gdal_translate`` to crop the scene, use the argument
        ``-co SCHEME=BIL`` to make the output

    .. note ::

        Data are in radians but no transformation to
        meters yet, as ``wavelength`` or at least sensor name is not
        provided in the XML file.
    """

    def validate(self, filename, **kwargs):
        try:
            self._getDisplacementFile(filename)
            self._getLOSFile(filename)
            return True
        except ImportError:
            return False

    def _getLOSFile(self, path):
        if not op.isdir(path):
            path = op.dirname(path)
        rdr_files = glob.glob(op.join(path, "*.rdr.geo"))

        if len(rdr_files) == 0:
            raise ImportError("Could not find LOS file (*.rdr.geo)")

        rdr_file = rdr_files[0]
        self._log.info("Found LOS file: %s", rdr_file)
        return rdr_file

    def _getDisplacementFile(self, path):
        if op.isfile(path):
            disp_file = path
        else:
            files = glob.glob(op.join(path, "*.unw.geo"))
            if len(files) == 0:
                raise ImportError(
                    "Could not find displacement file " "(.unw.geo) at %s", path
                )
            disp_file = files[0]

        if not op.isfile("%s.xml" % disp_file):
            raise ImportError(
                "Could not find displacement XML file "
                "(%s.unw.geo.xml)" % op.basename(disp_file)
            )
        self._log.info("Found Displacement file: %s", disp_file)
        return disp_file

    def read(self, path, **kwargs):
        path = op.abspath(path)
        container = self.container

        xml_file = self._getDisplacementFile(path) + ".xml"
        self._log.info("Parsing ISCE XML file %s" % xml_file)
        isce_xml = ISCEXMLParser(xml_file)

        coord_lon = isce_xml.getProperty("coordinate1")
        coord_lat = isce_xml.getProperty("coordinate2")
        container.frame.dN = np.abs(coord_lat["delta"])
        container.frame.dE = np.abs(coord_lon["delta"])

        nlon = int(coord_lon["size"])
        nlat = int(coord_lat["size"])

        container.frame.spacing = "degree"
        container.frame.llLat = coord_lat["startingvalue"] + (nlat * coord_lat["delta"])
        container.frame.llLon = coord_lon["startingvalue"]

        displ = np.memmap(self._getDisplacementFile(path), dtype="<f4").reshape(
            nlat, nlon * 2
        )[:, nlon:]

        displ = np.flipud(displ)
        displ[displ == 0.0] = np.nan
        container.displacement = displ

        los_file = self._getLOSFile(path)
        los_data = np.fromfile(los_file, dtype="<f4").reshape(nlat * 2, nlon)

        theta = np.flipud(los_data[0::2, :])
        phi = np.flipud(los_data[1::2, :])

        def los_is_degree():
            return np.abs(theta).max() > np.pi or np.abs(phi).max() > np.pi

        if not los_is_degree():
            raise ImportError(
                "The LOS file (%s) seems to be in radians! "
                "Change it to degree!" % op.basename(los_file)
            )

        phi[phi == 0.0] = np.nan
        theta[theta == 0.0] = np.nan

        phi *= d2r
        theta *= d2r

        phi = np.pi / 2 + phi
        theta = np.pi / 2 - theta

        container.phi = phi
        container.theta = theta

        return container


class GMTSAR(SceneIO):
    """
    Reading GMTSAR grid files.

    .. note ::

        Expects:

        * Displacement grid (NetCDF, :file:`*los_ll.grd`) in meter
          (in case use "gmt grdmath los_cm_ll.grd 0.01 MUL = los_m_ll.grd')
        * LOS binary data (see instruction, :file:`*.los.enu`)

    Calculate the corresponding unit look vectors with GMT5SAR ``SAT_look``:

    .. code-block:: sh

        gmt grd2xyz los_ll.grd | gmt grdtrack -Gdem.grd | \\
        awk {'print $1, $2, $4'} | \\
        SAT_look 20050731.PRM -bos > 20050731.los.enu
    """

    def validate(self, filename, **kwargs):
        try:
            if self._getDisplacementFile(filename)[-4:] == ".grd":
                return True
        except ImportError:
            return False
        return False

    def _getLOSFile(self, path):
        if not op.isdir(path):
            path = op.dirname(path)
        los_files = glob.glob(op.join(path, "*.los.*"))
        if len(los_files) == 0:
            self._log.warning(GMTSAR.__doc__)
            raise ImportError("Could not find LOS file (*.los.*)")
        los_file = los_files[0]
        self._log.debug("Found LOS file: %s", los_file)
        return los_file

    def _getDisplacementFile(self, path):
        if op.isfile(path):
            return path
        else:
            files = glob.glob(op.join(path, "*.grd"))
            if len(files) == 0:
                raise ImportError(
                    "Could not find displacement file " "(*.grd) at %s", path
                )
            disp_file = files[0]
        self._log.debug("Found Displacement file: %s", disp_file)
        return disp_file

    def read(self, path, **kwargs):
        from scipy.io import netcdf_file

        path = op.abspath(path)
        container = self.container

        grd = netcdf_file(self._getDisplacementFile(path), mode="r", version=2)
        displ = grd.variables["z"][:].copy()
        container.displacement = displ
        shape = container.displacement.shape
        # LatLon
        container.frame.spacing = "degree"
        container.frame.llLat = grd.variables["lat"][:].min()
        container.frame.llLon = grd.variables["lon"][:].min()

        container.frame.dN = (
            grd.variables["lat"][:].max() - container.frame.llLat
        ) / shape[0]
        container.frame.dE = (
            grd.variables["lon"][:].max() - container.frame.llLon
        ) / shape[1]

        # Theta and Phi
        try:
            los = np.memmap(self._getLOSFile(path), dtype="<f4")
            e = los[3::6].copy().reshape(shape)
            n = los[4::6].copy().reshape(shape)
            u = los[5::6].copy().reshape(shape)

            phi = np.arctan(n / e)
            theta = np.arcsin(u)
            # phi[n < 0] += np.pi

            container.phi = phi
            container.theta = theta
        except ImportError:
            self._log.warning(self.__doc__)
            self._log.warning("Defaulting theta to pi/2 and phi to 0.")
            container.theta = np.pi / 2
            container.phi = 0.0
        return container


class SARscape(SceneIO):
    """
    Reading SARscape :file:`*_disp` files.

    .. note ::

        Expects:

        * Header file in :file:`*_disp.hdr`
        * Displacement data in cm in :file:`*_disp`
        * LOS data in :file:`*disp_ILOS` and :file:`*disp_ALOS` files.
    """

    def read(self, filename, **kwargs):
        header = self.parseHeaderFile(filename)

        def load_data(filename):
            self._log.debug("Loading %s" % filename)
            return np.flipud(
                np.fromfile(filename, dtype=np.float32).reshape(
                    (header.lines, header.samples)
                )
            )

        displacement = load_data(filename)
        theta_file, phi_file = self.getLOSFiles(filename)

        if not theta_file:
            theta = np.full_like(displacement, 0.0)
        else:
            theta = load_data(theta_file)
            theta = np.deg2rad(theta)

        if not phi_file:
            phi = np.full_like(displacement, np.pi / 2)
        else:
            phi = load_data(phi_file)
            phi = np.pi / 2 - np.rad2deg(phi)

        container = self.container
        container.displacement = displacement
        container.phi = phi
        container.theta = theta

        map_info = header.map_info
        container.frame.dE = float(map_info[5])
        container.frame.dN = dN = float(map_info[6])
        container.frame.spacing = "meter"

        container.frame.llLat, container.frame.llLon = utm.to_latlon(
            float(map_info[3]),
            float(map_info[4]) - header.lines * dN,
            zone_number=int(map_info[7]),
            northern=True if map_info[8] == "Northern" else False,
        )

        return container

    def parseHeaderFile(self, filename):
        hdr_file = self._getHDRFile(filename)
        conf = re.compile(r"^(.+)\s+=\s+(.+)\n", re.MULTILINE)

        header = AttribDict()
        with open(hdr_file) as f:
            s = f.read()

            linebreaks = re.compile(r"{(.+)\n?(.+)}")
            s = linebreaks.sub(r"{ \g<1> \g<2> }", s)

            for match in conf.finditer(s):
                groups = match.groups()
                key = groups[0].strip().replace(" ", "_")
                value = groups[1].strip()
                try:
                    value = int(value)
                except ValueError:
                    pass

                header[key] = value

            header.map_info = header.map_info.strip("{} ").split(", ")
            if not len(header.map_info) == 11:
                raise HeaderError("`map info` header is not consistent!")
            if header.map_info[0] != "UTM":
                raise HeaderError("`map info` is not UTM!")

        return header

    def getLOSFiles(self, filename):
        ilos_file = op.abspath(filename + "_ILOS")
        if not op.exists(ilos_file):
            self._log.warning("Could not find ILOS file! (%s)" % ilos_file)
            ilos_file = False

        alos_file = op.abspath(filename + "_ALOS")
        if not op.exists(alos_file):
            self._log.warning("Could not find ALSO file! (%s)" % alos_file)
            alos_file = False
        return ilos_file, alos_file

    def _getHDRFile(self, filename):
        hdr_file = op.abspath(op.splitext(filename)[0] + ".hdr")
        if not op.exists(hdr_file):
            raise OSError("SARscape .hdr file not found (%s)" % hdr_file)
        return hdr_file

    def validate(self, filename, **kwargs):
        val = re.compile(r"SARscape|ENVI Standard", re.MULTILINE)
        try:
            hdr_file = self._getHDRFile(filename)
        except OSError:
            return False

        with open(hdr_file) as f:
            res = val.search(f.read())
            if res is not None:
                return True
            return False


class LiCSAR(SceneIO):
    """
    Import unwrapped Geotiffs from the
    `COMET LiCSAR Portal <https://comet.nerc.ac.uk/COMET-LiCS-portal/>`_.

    .. note ::

        Requires the Python package
        `gdal/osgeo <https://pypi.org/project/GDAL/>`_! Or through

        Expects:

        * Unwrapped geotiff in :file:`*.unw.tif`
        * LOS data in :file:`*.geo.[NEU].tif` files

    See also the download script in :mod:`kite.clients`.
    """

    def _getLOS(self, filename, component):
        path = op.dirname(filename)
        fn = glob.glob(op.join(path, component))
        if len(fn) != 1:
            raise ImportError("Cannot find LOS vector file %s!" % component)

        dataset = gdal.Open(fn[0], gdal.GA_ReadOnly)
        return self._readBandData(dataset)

    @staticmethod
    def _readBandData(dataset, band=1):
        band = dataset.GetRasterBand(band)
        array = band.ReadAsArray()
        array[array == band.GetNoDataValue()] = np.nan

        return np.flipud(array)

    def read(self, filename, **kwargs):
        dataset = gdal.Open(filename, gdal.GA_ReadOnly)
        georef = dataset.GetGeoTransform()

        llLon = georef[0]
        llLat = georef[3] + dataset.RasterYSize * georef[5]

        c = self.container

        c.frame.spacing = "degree"
        c.frame.llLat = llLat
        c.frame.llLon = llLon
        c.frame.dE = georef[1]
        c.frame.dN = abs(georef[5])

        displacement = self._readBandData(dataset)
        c.displacement = -displacement / (4 * np.pi) * LAMBDA_SENTINEL

        try:
            los_n = self._getLOS(filename, "*.geo.N.tif")
            los_e = self._getLOS(filename, "*.geo.E.tif")
            los_u = self._getLOS(filename, "*.geo.U.tif")
        except ImportError:
            self._log.warning(
                "Cannot find LOS angle files *.geo.[NEU].tif,"
                " using static Sentinel-1 descending angles."
            )

            heading = 83.0
            incident = 50.0

            un = np.sin(d2r * incident) * np.cos(d2r * heading)
            ue = np.sin(d2r * incident) * np.sin(d2r * heading)
            uz = np.cos(d2r * incident)

            los_n = np.full_like(c.displacement, un)
            los_e = np.full_like(c.displacement, ue)
            los_u = np.full_like(c.displacement, uz)

        c.phi = np.arctan2(los_n, los_e)
        c.theta = np.arcsin(los_u)

        c.meta.title = dataset.GetDescription()

        return c

    def validate(self, filename, **kwargs):
        if gdal.IdentifyDriver(filename) is None:
            return False
        return True


class ARIA(SceneIO):
    """
    Import unwrapped InSAR scenes from the
    `NASA/JPL ARIA <https://aria.jpl.nasa.gov/>`_ GUNW data products.

    .. note ::

        Requires the Python package
        `gdal/osgeo <https://pypi.org/project/GDAL/>`_! Or through

        Expects:

        * Extracted layers: unwrappedPhase, lookAngle, incidenceAngle,
          connectedComponents

    Use ``ariaExtract.py`` to extract the layers:

    .. code-block:: sh

        ariaExtract.py -w ascending -f aria-data.nc -d download \\
        -l unwrappedPhase,incidenceAngle,lookAngle

    """

    @staticmethod
    def _readBandData(dataset, band=1):
        band = dataset.GetRasterBand(band)
        array = band.ReadAsArray()
        if array.dtype != np.int16 and array.dtype != np.int:
            array[array == band.GetNoDataValue()] = np.nan

        return np.flipud(array)

    @staticmethod
    def _dataset_from_dir(folder):
        files = set(f for f in os.scandir(folder) if f.is_file())
        for f in files:
            if op.splitext(f.name)[-1] == "":
                break
        else:
            raise ImportError("could not load dataset from %s" % folder)

        return gdal.Open(f.path, gdal.GA_ReadOnly)

    def read(self, folder, **kwargs):
        unw_phase = self._dataset_from_dir(op.join(folder, "unwrappedPhase"))
        georef = unw_phase.GetGeoTransform()

        llLon = georef[0]
        llLat = georef[3] + unw_phase.RasterYSize * georef[5]

        c = self.container

        c.frame.spacing = "degree"
        c.frame.llLat = llLat
        c.frame.llLon = llLon
        c.frame.dE = georef[1]
        c.frame.dN = abs(georef[5])

        conn_comp = self._dataset_from_dir(op.join(folder, "connectedComponents"))

        displacement = self._readBandData(unw_phase)
        conn_mask = self._readBandData(conn_comp)  # Mask from snaphu
        displacement *= np.where(conn_mask, 1.0, np.nan)

        c.displacement = displacement / (4 * np.pi) * LAMBDA_SENTINEL

        inc_angle = self._dataset_from_dir(op.join(folder, "incidenceAngle"))
        azi_angle = self._dataset_from_dir(op.join(folder, "azimuthAngle"))

        c.theta = np.pi / 2 - self._readBandData(inc_angle) * d2r
        c.phi = self._readBandData(azi_angle) * d2r

        c.meta.scene_id = op.basename(unw_phase.GetDescription())
        c.meta.scene_title = c.meta.scene_id

        t_slave, t_master = c.meta.scene_id.split("_")
        c.meta.time_master = datetime(
            *time.strptime(t_master, "%Y%m%d")[:6]
        ).timestamp()
        c.meta.time_slave = datetime(*time.strptime(t_slave, "%Y%m%d")[:6]).timestamp()

        c.meta.satellite_name = "undefined (ARIA)"

        return c

    def validate(self, folder, **kwargs):
        expected_dirs = set(
            ["unwrappedPhase", "incidenceAngle", "lookAngle", "connectedComponents"]
        )
        if not op.isdir(folder):
            return False

        dirs = set(d.name for d in os.scandir(folder) if d.is_dir())
        if not expected_dirs - dirs:
            return True

        return False


class SNAP_Gamma(SceneIO):
    """SNAP import

    Reading geocoded displacement maps (unwrapped igs) originating
    from SNAP software using the export option.

    When georeferencing the scene, export the incidence angle from ellpsoid in
    the tab **Processing Parameters**.
    Export the unwrapped **displacement** band and
    **incidenceAngleFromEllipsoid** to Gamma format:

    1. File -> Export -> SAR Formats -> Gamma

    Select **Metadata -> Abstracted_Metadata** and
    2. File -> Export -> Other -> Product Metadata.

    .. note :: Expects:

        * [:file:`*`] Binary file from SNAP Gamma Export with displacement in
            radians
        * [:file:`*.Abstracted_Metadata.txt`] Metadata (parameter) file
            If you want to translate radians to
            meters using the `radar_frequency`.
        * [:file:`*par`] Parameter file, describing ``first_near_lat,
        last_near_long, num_output_lines, num_samples_per_line,
                lat_pixel_res, lon_pixel_res, radar_frequency and heading.``
        * [:file:`incidenceAngleFromEllipsoid.rslc`] Incidence angle file.

    .. warning ::

        * Data has to be georeferenced to latitude/longitude or UTM!
        * Look vector files - expected to have a particular name
    """

    @staticmethod
    def _parseParameterFile(filename):
        params = {}
        rc = re.compile(r"([\w]*)\s*([\w.+-]*)")
        with open(filename, mode="r") as par:
            for line in par:
                if line[:29] == "metadata.Abstracted_Metadata.":
                    parsed = rc.match(line[29:])
                    if parsed is None:
                        continue

                    groups = parsed.groups()
                    params[groups[0]] = safe_cast(
                        groups[1], float, default=groups[1].strip()
                    )
        return params

    def _getParameters(self, path, log=False):
        required_utm = (
            "post_north",
            "post_east",
            "corner_east",
            "corner_north",
            "num_output_lines",
            "num_samples_per_line",
        )
        required_lat_lon = (
            "first_near_lat",
            "last_near_long",
            "num_output_lines",
            "num_samples_per_line",
            "lat_pixel_res",
            "lon_pixel_res",
        )

        par_file = glob.glob(op.join(path, "*Abstracted_Metadata.txt"))
        for file in par_file:
            params = self._parseParameterFile(file)
            if check_required(required_utm, params) or check_required(
                required_lat_lon, params
            ):
                if not log:
                    self._log.info("Found parameter file %s" % file)
                return op.basename(file), params

        raise ImportError(
            "Parameter file %s does not hold required parameters" % par_file
        )

    def validate(self, filename, **kwargs):
        try:
            folder = op.dirname(op.abspath(filename))
            self._getParameters(folder)
            return True
        except ImportError:
            return False

    def _getLOSAngles(self, filename, pattern):
        path = op.dirname(op.realpath(filename))
        phi_files = glob.glob(op.join(path, pattern))
        if len(phi_files) == 0:
            self._log.warning(
                "Could not find LOS file %s, " "defaulting to angle to 0.", pattern
            )
            return 0.0
        elif len(phi_files) > 1:
            self._log.warning(
                "Found multiple LOS files %s, " "defaulting to angle 0.", pattern
            )
            return 0.0

        filename = phi_files[0]
        self._log.info("Loading LOS %s from %s" % (pattern, filename))
        return np.memmap(filename, mode="r", dtype=">f4")

    def read(self, filename, **kwargs):
        """
        :param filename: Gamma software parameter file
        :type filename: str
        :param par_file: Corresponding parameter (:file:`*par`) file.
                         (optional)
        :type par_file: str
        :returns: Import dictionary
        :rtype: dict
        :raises: ImportError
        """
        par_file, params = self._getParameters(
            op.dirname(op.abspath(filename)), log=True
        )

        ncols = int(params["num_samples_per_line"])
        nlines = int(params["num_output_lines"])
        radar_frequency = params.get("radar_frequency", None)
        heading_par = float(params.get("centre_heading", None))
        displ = np.fromfile(filename, dtype=">f4")

        # Resize array if last line is not scanned completely
        fill = 0
        if (displ.size % ncols) != 0:
            fill = np.empty(ncols - displ.size % ncols)
            fill.fill(np.nan)
            displ = np.append(displ, fill)

        displ = displ.reshape(nlines, ncols)
        displ[displ == -0.0] = np.nan
        displ = np.flipud(displ)

        if radar_frequency and "_dsp_" not in par_file:
            radar_frequency = float(radar_frequency)
            radar_frequency *= 1e6  # SNAP gives MHz
            self._log.info(
                "Scaling displacement by radar_frequency %f GHz", radar_frequency / 1e9
            )
            wavelength = util.C / radar_frequency
            displ /= -4 * np.pi
            displ *= wavelength

        elif not radar_frequency and "_dsp_" not in par_file:
            self._log.warning("Could not determine radar_frequency!")
            wavelength = None

        else:
            wavelength = None

        inc_angle = self._getLOSAngles(
            filename, "incidenceAngleFromEllipsoid.rslc"
        ).copy()
        if fill:
            inc_angle = np.append(inc_angle, fill)
        inc_angle[inc_angle == 0.0] = np.nan

        phi = np.full_like(displ, (180.0 - heading_par))
        theta = 90.0 - inc_angle.reshape(displ.shape)
        theta = np.flipud(theta)

        c = self.container

        c.displacement = displ
        c.theta = theta * d2r
        c.phi = phi * d2r

        c.meta.wavelength = wavelength
        c.meta.title = params.get("PRODUCT", "SNAP Import")
        c.meta.satellite_name = params.get("SPH_DESCRIPTOR", "None")

        orb = params.get("PASS", None)
        c.meta.orbital_node = orb.title() if orb else None

        c.bin_file = filename
        c.par_file = par_file

        if params["map_projection"] == "UTM":
            utm_zone = params["projection_zone"]
            try:
                utm_zone_letter = utm.latitude_to_zone_letter(params["center_latitude"])
            except ValueError:
                self._log.warning(
                    "Could not parse UTM Zone letter," " defaulting to N!"
                )
                utm_zone_letter = "N"

            self._log.info("Using UTM reference: Zone %d%s", utm_zone, utm_zone_letter)
            c.frame.spacing = "meter"

            dN = abs(params["post_north"])
            dE = abs(params["post_east"])

            utm_corn_e = params["corner_east"]
            utm_corn_n = params["corner_north"]

            utm_corn_eo = utm_corn_e + dE * displ.shape[1]
            utm_corn_no = utm_corn_n + dN * displ.shape[0]

            utm_e = np.linspace(utm_corn_e, utm_corn_eo, displ.shape[1])
            utm_n = np.linspace(utm_corn_n, utm_corn_no, displ.shape[0])

            llLat, llLon = utm.to_latlon(
                utm_e.min(), utm_n.min(), utm_zone, utm_zone_letter
            )

        else:
            self._log.info("Using Lat/Lon reference")
            c.frame.spacing = "degree"

            if orb.lower() == "ascending":
                llLat = params["last_near_lat"]
                llLon = params["last_near_long"]

            elif orb.lower() == "descending":
                llLat = params["last_far_lat"]
                llLon = params["first_near_long"]

            else:
                raise AttributeError("cannot determine orbit")

            dE = abs(params["lon_pixel_res"])
            dN = abs(params["lat_pixel_res"])

        c.frame.llLat = llLat
        c.frame.llLon = llLon

        c.frame.dE = dE
        c.frame.dN = dN

        return c



import logging
import os
from os import path

from PyQt5 import QtCore, QtGui, QtWidgets
from pyrocko.guts import Bool, Float, Int, List, Object, String, Tuple, load

from kite.qt_utils import loadUi

from .util import get_resource

config_file = path.expanduser("~/.config/kite/talpa_config.yml")
logger = logging.getLogger("TalpaConfig")

config_instance = None


class TalpaConfig(Object):
    show_cursor = Bool.T(default=True)

    default_gf_dir = String.T(default="")  # noqa

    nvectors = Int.T(
        default=200, help="Number of horizontal displacement vectors to show"
    )

    vector_color = Tuple.T(
        default=(0, 0, 0, 100),
        help="Color of the displacement arrows, RGBA (0, 0, 0, 100)",
    )

    vector_relative_length = Int.T(default=100, help="Relative length of the arrow.")

    vector_pen_thickness = Float.T(default=1.0, help="Thickness of the arrows.")

    view_north = Bool.T(default=True, help="Show the north view of displacement.")

    view_east = Bool.T(default=True, help="Show the east view of displacement.")

    view_down = Bool.T(default=True, help="Show the down view of displacement.")

    view_los = Bool.T(default=True, help="Show the los view of displacement.")

    def __init__(self, *args, **kwargs):
        class QConfig(QtCore.QObject):
            updated = QtCore.pyqtSignal()

        Object.__init__(self, *args, **kwargs)
        self.qconfig = QConfig()

    def saveConfig(self):
        self.regularize()
        self.dump(filename=config_file)
        self.qconfig.updated.emit()


def createDefaultConfig():
    logger.info("Creating new config...")
    import os

    try:
        os.makedirs(path.dirname(config_file))
    except OSError as e:
        if e.errno == 17:
            pass
        else:
            raise e

    config = TalpaConfig()
    config.dump(filename=config_file)


def getConfig():
    global config_instance

    if not config_instance:
        if not path.isfile(config_file):
            createDefaultConfig()
        try:
            logger.info("Loading config from %s..." % config_file)
            config_instance = load(filename=config_file)
        except KeyError:
            createDefaultConfig()
            config_instance = TalpaConfig()

    return config_instance


class ConfigDialog(QtWidgets.QDialog):

    attributes = [
        "show_cursor",
        "default_gf_dir",
        "nvectors",
        "vector_color",
        "vector_relative_length",
        "vector_pen_thickness",
        "view_east",
        "view_north",
        "view_down",
        "view_los",
    ]

    def __init__(self, *args, **kwargs):
        QtWidgets.QDialog.__init__(self, *args, **kwargs)

        self.completer = QtWidgets.QCompleter()
        self.completer_model = QtWidgets.QFileSystemModel(self.completer)
        self.completer.setModel(self.completer_model)
        self.completer.setMaxVisibleItems(8)

        loadUi(get_resource("dialog_config.ui"), self)

        self.ok_button.released.connect(self.setAttributes)
        self.ok_button.released.connect(self.close)

        self.apply_button.released.connect(self.setAttributes)

        self.vector_color_picker = QtWidgets.QColorDialog(self)
        self.vector_color_picker.setCurrentColor(
            QtGui.QColor(*getConfig().vector_color)
        )
        self.vector_color_picker.setOption(self.vector_color_picker.ShowAlphaChannel)
        self.vector_color_picker.colorSelected.connect(self.updateVectorColor)
        self.vector_color_picker.setModal(True)
        self.vector_color.clicked.connect(self.vector_color_picker.show)

        self.vector_color.setValue = self.setButtonColor
        self.vector_color.value = self.getButtonColor

        self.chooseStoreDirButton.released.connect(self.chooseStoreDir)
        self.completer_model.setRootPath("")
        self.completer.setParent(self.default_gf_dir)
        self.default_gf_dir.setCompleter(self.completer)

        self.getAttributes()

    @QtCore.pyqtSlot()
    def chooseStoreDir(self):
        folder = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Open Pyrocko GF Store", os.getcwd()
        )
        if folder != "":
            self.default_gf_dir.setText(folder)

        self.setAttributes()

    def getAttributes(self):
        for attr in self.attributes:
            qw = self.__getattribute__(attr)
            value = getConfig().__getattribute__(attr)
            if isinstance(value, bool):
                qw.setChecked(value)
            elif isinstance(value, str):
                qw.setText(value)
            else:
                qw.setValue(value)

    @QtCore.pyqtSlot()
    def setAttributes(self):
        for attr in self.attributes:
            qw = self.__getattribute__(attr)
            if isinstance(qw, QtWidgets.QCheckBox):
                value = qw.isChecked()
            elif isinstance(qw, QtWidgets.QLineEdit):
                value = str(qw.text())
            else:
                value = qw.value()

            getConfig().__setattr__(attr, value)

        getConfig().saveConfig()

    def setButtonColor(self, rgba):
        self.vector_color.setStyleSheet(
            "background-color: rgb(%d, %d, %d, %d);" "border: none;" % rgba
        )

    def getButtonColor(self):
        return getConfig().vector_color

    @QtCore.pyqtSlot(QtGui.QColor)
    def updateVectorColor(self, qcolor):
        getConfig().vector_color = (
            qcolor.red(),
            qcolor.green(),
            qcolor.blue(),
            qcolor.alpha(),
        )
        self.setButtonColor(getConfig().vector_color)



import shutil
import tempfile
import unittest

import numpy as np

from kite import SandboxScene, TestSandboxScene  # noqa
from kite.sources import (
    PyrockoDoubleCouple,
    PyrockoMomentTensor,
    PyrockoRectangularSource,
    PyrockoRingfaultSource,
)

from . import common

plot = False
gf_store = "/home/marius/Development/testing/leeds/aquila_example/insar/gf_abruzzo_nearfield"  # noqa

common.setLogLevel("DEBUG")


class testSourcePyrocko(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ms = SandboxScene()
        cls.tmpdir = tempfile.mkdtemp(prefix="kite")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)

    def test_rectangular_source(self):
        nsources = 2

        def r(lo, hi):
            return np.random.randint(lo, high=hi, size=1).astype(float)

        for s in range(nsources):
            length = r(5000, 15000)
            self.ms.addSource(
                PyrockoRectangularSource(
                    easting=r(0.0, self.ms.frame.E.max()),  # ok
                    northing=r(0.0, self.ms.frame.N.max()),  # ok
                    depth=r(0, 8000),  # ok
                    strike=r(0, 360),  # ok
                    dip=r(0, 170),
                    slip=r(1, 7),  # ok
                    rake=r(0, 180),
                    length=length,
                    width=15.0 * length**0.66,
                    store_dir=gf_store,
                )
            )  # noqa

        self._plot_displacement(self.ms)

    def test_moment_tensor(self):
        nsources = 5

        def r(lo, hi):
            return np.random.randint(lo, high=hi, size=1).astype(float)

        for s in range(nsources):
            self.ms.addSource(
                PyrockoMomentTensor(
                    easting=r(0.0, self.ms.frame.E.max()),  # ok
                    northing=r(0.0, self.ms.frame.N.max()),  # ok
                    depth=r(0, 8000),  # ok
                    store_dir=gf_store,
                )
            )

        self._plot_displacement(self.ms)

    def test_double_couple(self):
        nsources = 5

        def r(lo, hi):
            return np.random.randint(lo, high=hi, size=1).astype(float)

        for s in range(nsources):
            self.ms.addSource(
                PyrockoDoubleCouple(
                    easting=r(0.0, self.ms.frame.E.max()),  # ok
                    northing=r(0.0, self.ms.frame.N.max()),  # ok
                    depth=r(0, 8000),  # ok
                    store_dir=gf_store,
                )
            )

        self._plot_displacement(self.ms)

    def test_ring_fault(self):
        nsources = 1

        def r(lo, hi):
            return np.random.randint(lo, high=hi, size=1).astype(float)

        for s in range(nsources):
            diameter = r(5000, 15000)
            self.ms.addSource(
                PyrockoRingfaultSource(
                    easting=r(0.0, self.ms.frame.E.max()),  # ok
                    northing=r(0.0, self.ms.frame.N.max()),  # ok
                    depth=r(0, 8000),  # ok
                    strike=r(0, 360),  # ok
                    dip=r(0, 170),
                    magnitude=r(2, 6),  # ok
                    diameter=diameter,
                    store_dir=gf_store,
                )
            )  # noqa

        self._plot_displacement(self.ms)

    @staticmethod
    def _plot_displacement(ms):
        if not plot:
            ms.down
            return

        import matplotlib.pyplot as plt
        from matplotlib.patches import Polygon  # noqa

        fig = plt.figure()
        ax = fig.gca()
        ms.processSources()

        ax.imshow(
            np.flipud(ms.north),
            aspect="equal",
            extent=[0, ms.frame.E.max(), 0, ms.frame.N.max()],
        )
        # for src in ms.sources:
        #     for seg in src.segments:
        #         p = Polygon(seg.outline(), alpha=.8, fill=False)
        #         ax.add_artist(p)
        plt.show()
        fig.clear()


if __name__ == "__main__":
    plot = True
    unittest.main(exit=False)



import os

import numpy as np
from PyQt5 import QtCore, QtGui

from kite.sources import (
    PyrockoDoubleCouple,
    PyrockoMomentTensor,
    PyrockoRectangularSource,
    PyrockoRingfaultSource,
    PyrockoVLVDSource,
)

from ..config import getConfig
from .base import PointSourceROI, RectangularSourceROI, SourceDelegate, SourceEditDialog

d2r = np.pi / 180.0
r2d = 180.0 / np.pi


class PyrockoSourceDialog(SourceEditDialog):
    def __init__(self, delegate, ui_file, *args, **kwargs):
        SourceEditDialog.__init__(self, delegate, ui_file, *args, **kwargs)

        self.completer = QtGui.QCompleter()
        self.completer_model = QtGui.QFileSystemModel(self.completer)
        self.completer.setModel(self.completer_model)
        self.completer.setMaxVisibleItems(8)

        self.chooseStoreDirButton.released.connect(self.chooseStoreDir)

        self.completer_model.setRootPath("")
        self.completer.setParent(self.store_dir)
        self.store_dir.setCompleter(self.completer)

        self.store_dir.setValue = self.store_dir.setText
        self.store_dir.value = self.store_dir.text

        self.getSourceParameters()

    @QtCore.pyqtSlot()
    def chooseStoreDir(self):
        folder = QtGui.QFileDialog.getExistingDirectory(
            self, "Open Pyrocko GF Store", os.getcwd()
        )
        if folder != "":
            self.store_dir.setText(folder)


class PyrockoRectangularSourceDelegate(SourceDelegate):

    __represents__ = "PyrockoRectangularSource"

    display_backend = "pyrocko"
    display_name = "RectangularSource"

    parameters = [
        "easting",
        "northing",
        "width",
        "length",
        "depth",
        "slip",
        "strike",
        "dip",
        "rake",
        "store_dir",
        "decimation_factor",
    ]
    ro_parameters = []

    class RectangularSourceDialog(PyrockoSourceDialog):
        def __init__(self, *args, **kwargs):
            PyrockoSourceDialog.__init__(
                self, ui_file="pyrocko_rectangular_source.ui", *args, **kwargs
            )

    ROIWidget = RectangularSourceROI
    EditDialog = RectangularSourceDialog

    @staticmethod
    def getRepresentedSource(sandbox):
        length = 5000.0
        return PyrockoRectangularSource(
            depth=4000,
            length=length,
            width=15.0 * length**0.66,
            strike=45.0,
            rake=0,
            slip=2,
            store_dir=getConfig().default_gf_dir or "",
        )

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>Width:</td><td>{source.width:.2f} m</td>
</tr><tr>
    <td>Length:</td><td>{source.length:.2f} m</td>
</tr><tr>
    <td>Strike:</td><td>{source.strike:.2f}&deg;</td>
</tr><tr>
    <td>Dip:</td><td>{source.dip:.2f}&deg;</td>
</tr><tr>
    <td>Rake:</td><td>{source.rake:.2f}&deg;</td>
</tr><tr style="font-weight: bold;">
    <td>Slip:</td><td>{source.slip:.2f} m</td>
</tr></table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)


class PyrockoMomentTensorDelegate(SourceDelegate):

    __represents__ = "PyrockoMomentTensor"

    display_backend = "pyrocko"
    display_name = "MomentTensor"

    parameters = [
        "easting",
        "northing",
        "depth",
        "store_dir",
        "mnn",
        "mee",
        "mdd",
        "mne",
        "mnd",
        "med",
    ]
    ro_parameters = []

    class MomentTensorDialog(PyrockoSourceDialog):

        scaling_params = ["mnn", "mee", "mdd", "mne", "mnd", "med"]

        def __init__(self, *args, **kwargs):
            PyrockoSourceDialog.__init__(
                self, ui_file="pyrocko_moment_tensor.ui", *args, **kwargs
            )

        @QtCore.pyqtSlot()
        def setSourceParameters(self):
            params = {}
            scale = float("1e%d" % self.exponent.value())
            for param in self.delegate.parameters:
                params[param] = self.__getattribute__(param).value()
                if param in self.scaling_params:
                    params[param] = params[param] * scale
            self.delegate.updateModelParameters(params)

        @QtCore.pyqtSlot()
        def getSourceParameters(self):
            params = self.delegate.getSourceParameters()
            exponent = np.log10(
                np.max([v for k, v in params.items() if k in self.scaling_params])
            )
            scale = float("1e%d" % int(exponent))

            for param, value in params.items():
                if param in self.scaling_params:
                    self.__getattribute__(param).setValue(value / scale)
                else:
                    self.__getattribute__(param).setValue(value)

    ROIWidget = PointSourceROI
    EditDialog = MomentTensorDialog

    @staticmethod
    def getRepresentedSource(sandbox):
        return PyrockoMomentTensor(store_dir=getConfig().default_gf_dir or "")

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>mnn:</td><td style="align: justify;">{source.mnn:.3e}</td><td>Nm</td>
</tr><tr>
    <td>mee:</td><td style="align: justify;">{source.mee:.3e}</td><td>Nm</td>
</tr><tr>
    <td>mdd:</td><td style="align: justify;">{source.mdd:.3e}</td><td>Nm</td>
</tr><tr>
    <td>mne:</td><td style="align: justify;">{source.mne:.3e}</td><td>Nm</td>
</tr><tr>
    <td>mnd:</td><td style="align: justify;">{source.mnd:.3e}</td><td>Nm</td>
</tr><tr>
    <td>med:</td><td style="align: justify;">{source.med:.3e}</td><td>Nm</td>
</tr>
</table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)


class PyrockoDoubleCoupleDelegate(SourceDelegate):

    __represents__ = "PyrockoDoubleCouple"

    display_backend = "pyrocko"
    display_name = "DoubleCouple"

    parameters = [
        "easting",
        "northing",
        "depth",
        "store_dir",
        "strike",
        "dip",
        "rake",
        "magnitude",
    ]
    ro_parameters = []

    class DoubleCoupleDialog(PyrockoSourceDialog):
        def __init__(self, *args, **kwargs):
            PyrockoSourceDialog.__init__(
                self, ui_file="pyrocko_double_couple.ui", *args, **kwargs
            )

    class DoubleCoupleROI(PointSourceROI):
        def __init__(self, *args, **kwargs):
            PointSourceROI.__init__(self, *args, **kwargs)
            self.addRotateHandle([0.5, 1.0], [0.5, 0.5])
            self.updateROIPosition()

        @QtCore.pyqtSlot()
        def setSourceParametersFromROI(self):
            angle = self.angle()
            strike = float(-angle) % 360
            vec_x, vec_y = self._vectorToCenter(strike)

            parameters = {
                "easting": float(self.pos().x() + vec_x),
                "northing": float(self.pos().y() + vec_y),
                "strike": strike,
            }

            self.newSourceParameters.emit(parameters)

        @QtCore.pyqtSlot()
        def updateROIPosition(self):
            source = self.source
            vec_x, vec_y = self._vectorToCenter(source.strike)

            self.setAngle(-source.strike, finish=False)
            self.setPos(
                QtCore.QPointF(source.easting - vec_x, source.northing - vec_y),
                finish=False,
            )
            # self.setPos(QtCore.QPointF(source.easting, source.northing),
            #             finish=False)

        def _vectorToCenter(self, angle):
            rangle = angle * d2r

            sdx = self.size().x() / 2
            sdy = self.size().y() / 2

            return (
                sdx * np.sin(rangle) + sdy * np.cos(rangle),
                sdx * np.cos(rangle) - sdy * np.sin(rangle),
            )

    ROIWidget = DoubleCoupleROI
    EditDialog = DoubleCoupleDialog

    def __init__(self, model, source, index):
        QtCore.QObject.__init__(self)
        self.source = source
        self.model = model
        self.index = index
        self.rois = []

        self.editing_dialog = None

        if model.selection_model is not None:
            self.setSelectionModel()

        self.model.selectionModelChanged.connect(self.setSelectionModel)

    @staticmethod
    def getRepresentedSource(sandbox):
        return PyrockoDoubleCouple(store_dir=getConfig().default_gf_dir or "")

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>Strike:</td><td>{source.strike:.2f}&deg;</td>
</tr><tr>
    <td>Dip:</td><td>{source.dip:.2f}&deg;</td>
</tr><tr>
    <td>Rake:</td><td>{source.rake:.2f}&deg;</td>
</tr><tr>
    <td>M<sub>0</sub>:</td><td>{source.moment:.2e}</td>
</tr><tr style="font-weight: bold;">
    <td>M<sub>W</sub>:</td><td>{source.magnitude:.2f}</td>
</tr>
</table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)


class PyrockoRingfaultDelegate(SourceDelegate):
    __represents__ = "PyrockoRingfaultSource"

    display_backend = "pyrocko"
    display_name = "Ringfault"

    parameters = [
        "store_dir",
        "easting",
        "northing",
        "depth",
        "diameter",
        "strike",
        "dip",
        "magnitude",
        "npointsources",
    ]
    ro_parameters = []

    class RingfaultDialog(PyrockoSourceDialog):
        def __init__(self, *args, **kwargs):
            PyrockoSourceDialog.__init__(
                self, ui_file="pyrocko_ringfault.ui", *args, **kwargs
            )

    class RingfaultROI(PointSourceROI):
        def __init__(self, *args, **kwargs):
            PointSourceROI.__init__(self, *args, **kwargs)
            self.addScaleRotateHandle([0.5, 1.0], [0.5, 0.5])
            self.updateROIPosition()

        @QtCore.pyqtSlot()
        def setSourceParametersFromROI(self):
            angle = self.angle()
            strike = float(-angle) % 360
            vec_x, vec_y = self._vectorToCenter(strike)

            parameters = {
                "easting": float(self.pos().x() + vec_x),
                "northing": float(self.pos().y() + vec_y),
                "diameter": self.size().y(),
                "strike": strike,
            }

            self.newSourceParameters.emit(parameters)

        @QtCore.pyqtSlot()
        def updateROIPosition(self):
            source = self.source

            self.setAngle(-source.strike, finish=False)
            self.setSize(source.diameter, finish=False)
            vec_x, vec_y = self._vectorToCenter(source.strike)
            self.setPos(
                QtCore.QPointF(source.easting - vec_x, source.northing - vec_y),
                finish=False,
            )
            # self.setPos(QtCore.QPointF(source.easting, source.northing),
            #             finish=False)

        def _vectorToCenter(self, angle):
            rangle = angle * d2r

            sdx = self.size().x() / 2
            sdy = self.size().y() / 2

            return (
                sdx * np.sin(rangle) + sdy * np.cos(rangle),
                sdx * np.cos(rangle) - sdy * np.sin(rangle),
            )

    EditDialog = RingfaultDialog
    ROIWidget = RingfaultROI

    def __init__(self, model, source, index):
        QtCore.QObject.__init__(self)
        self.source = source
        self.model = model
        self.index = index
        self.rois = []

        self.editing_dialog = None

        if model.selection_model is not None:
            self.setSelectionModel()

        self.model.selectionModelChanged.connect(self.setSelectionModel)

    @staticmethod
    def getRepresentedSource(sandbox):
        return PyrockoRingfaultSource(
            diameter=10000.0,
            store_dir=getConfig().default_gf_dir or "",
        )

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>Diameter:</td><td>{source.diameter:.2f} m</td>
</tr><tr>
    <td>Strike:</td><td>{source.strike:.2f}&deg;</td>
</tr><tr>
    <td>Dip:</td><td>{source.dip:.2f}&deg;</td>
</tr><tr style="font-weight: bold;">
    <td>M<sub>W</sub>:</td><td>{source.magnitude:.2f}</td>
</tr>
</table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)


class PyrockoVLVDSourceDelegate(SourceDelegate):

    __represents__ = "PyrockoVLVDSource"

    display_backend = "pyrocko"
    display_name = "VLVDSource"

    parameters = [
        "easting",
        "northing",
        "depth",
        "store_dir",
        "volume_change",
        "azimuth",
        "dip",
        "clvd_moment",
    ]
    ro_parameters = []

    class VLVDSourceDialog(PyrockoSourceDialog):

        scaling_params = ["clvd_moment"]

        def __init__(self, *args, **kwargs):
            PyrockoSourceDialog.__init__(
                self, ui_file="pyrocko_vlvd_source.ui", *args, **kwargs
            )

        @QtCore.pyqtSlot()
        def setSourceParameters(self):
            params = {}
            scale = float("1e%d" % self.clvd_moment_exponent.value())
            for param in self.delegate.parameters:
                params[param] = self.__getattribute__(param).value()
                if param in self.scaling_params:
                    params[param] = params[param] * scale
            self.delegate.updateModelParameters(params)

        @QtCore.pyqtSlot()
        def getSourceParameters(self):
            params = self.delegate.getSourceParameters()
            exponent = np.log10(
                np.max([abs(v) for k, v in params.items() if k in self.scaling_params])
            )
            scale = float("1e%d" % int(exponent))
            self.clvd_moment_exponent.setValue(int(exponent))

            for param, value in params.items():
                if param in self.scaling_params:
                    self.__getattribute__(param).setValue(value / scale)
                else:
                    self.__getattribute__(param).setValue(value)

    class VLVDSourceROI(PointSourceROI):
        def __init__(self, *args, **kwargs):
            PointSourceROI.__init__(self, *args, **kwargs)
            self.addRotateHandle([0.5, 1.0], [0.5, 0.5])
            self.updateROIPosition()

        @QtCore.pyqtSlot()
        def setSourceParametersFromROI(self):
            angle = self.angle()
            azimuth = float(-angle) % 360
            vec_x, vec_y = self._vectorToCenter(azimuth)

            parameters = {
                "easting": float(self.pos().x() + vec_x),
                "northing": float(self.pos().y() + vec_y),
                "azimuth": azimuth,
            }

            self.newSourceParameters.emit(parameters)

        @QtCore.pyqtSlot()
        def updateROIPosition(self):
            source = self.source
            vec_x, vec_y = self._vectorToCenter(source.azimuth)

            self.setAngle(-source.azimuth, finish=False)
            self.setPos(
                QtCore.QPointF(source.easting - vec_x, source.northing - vec_y),
                finish=False,
            )
            # self.setPos(QtCore.QPointF(source.easting, source.northing),
            #             finish=False)

        def _vectorToCenter(self, angle):
            rangle = angle * d2r

            sdx = self.size().x() / 2
            sdy = self.size().y() / 2

            return (
                sdx * np.sin(rangle) + sdy * np.cos(rangle),
                sdx * np.cos(rangle) - sdy * np.sin(rangle),
            )

    ROIWidget = VLVDSourceROI
    EditDialog = VLVDSourceDialog

    def __init__(self, model, source, index):
        QtCore.QObject.__init__(self)
        self.source = source
        self.model = model
        self.index = index
        self.rois = []

        self.editing_dialog = None

        if model.selection_model is not None:
            self.setSelectionModel()

        self.model.selectionModelChanged.connect(self.setSelectionModel)

    @staticmethod
    def getRepresentedSource(sandbox):
        return PyrockoVLVDSource(
            volume_change=0.25, store_dir=getConfig().default_gf_dir or ""
        )

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>Volume Change:</td><td>{source.volume_change:.2f} km<sup>3</sup></td>
</tr><tr>
    <td>CLVD Azimuth:</td><td>{source.azimuth:.2f}&deg;</td>
</tr><tr>
    <td>CLVD Dip:</td><td>{source.dip:.2f}&deg;</td>
</tr><tr style="font-weight: bold;">
    <td>CLVD M<sub>W</sub>:</td><td>{source.clvd_moment:.2e}</td>
</tr>
</table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)



import matplotlib.pyplot as plt
import numpy as num

from kite import SandboxScene
from kite.sources import OkadaSource, PyrockoRectangularSource

km = 1e3

sandbox = SandboxScene()
# Set the LOS incident angles, remember :class:`kite.Scene`
sandbox.phi.fill(num.rad2deg(100.0))
sandbox.theta.fill(num.rad2deg(23.0))

okada = OkadaSource(
    northing=40 * km,
    easting=40 * km,
    depth=4 * km,
    length=8 * km,
    width=4 * km,
    strike=63.0,
    dip=33.0,
    slip=3.0,
    opening=1,
)

pyrocko_rectangular = PyrockoRectangularSource(
    northing=40 * km,
    easting=40 * km,
    depth=4 * km,
    length=8 * km,
    width=4 * km,
    strike=63.0,
    dip=33.0,
    slip=3.0,
    store_dir="gfstore_halfspace",
)

sandbox.addSource(okada)

sandbox.processSources()

# Plot the resulting surface displacements
fig, axis = plt.subplots(nrows=2, ncols=2)
axis[0][0].imshow(sandbox.north)
axis[0][1].imshow(sandbox.east)
axis[1][0].imshow(sandbox.down)
axis[1][1].imshow(sandbox.displacement)  # Displacement in LOS
fig.show()



import numpy as np
import pyqtgraph as pg
import pyqtgraph.parametertree.parameterTypes as pTypes
from PyQt5 import QtCore, QtGui, QtWidgets
from pyqtgraph import dockarea

from kite.qt_utils import loadUi

from .base import KiteParameterGroup, KitePlot, KiteView, get_resource
from .tab_covariance import KiteSubplot

km = 1e3
pen_roi = pg.mkPen((78, 154, 6), width=2)
pen_roi_highlight = pg.mkPen((115, 210, 22), width=2, style=QtCore.Qt.DashLine)

pen_aps = pg.mkPen((255, 255, 255, 100), width=1.25)
brush_aps = pg.mkBrush((255, 255, 255, 100))
pen_aps_model = pg.mkPen((204, 0, 0), width=2, style=QtCore.Qt.DotLine)


class KiteAPS(KiteView):
    title = "Scene.APS"

    def __init__(self, spool):
        model = spool.model
        self.model = model

        self.main_widget = KiteAPSPlot(model)
        self.aps_correlation = KiteAPSCorrelation(self.main_widget)

        self.tools = {"APS Correlation": self.aps_correlation}

        self.aps_ctrl = EmpiricalAPSParams(model)
        self.parameters = [self.aps_ctrl]

        KiteView.__init__(self)

        for dock in self.tool_docks:
            dock.setStretch(10, 0.5)

        spool.actionApplyEmpiricalAPS.setChecked(model.getScene().aps.is_enabled())
        spool.actionApplyEmpiricalAPS.toggled.connect(self.toggleAPS)

        self.main_widget.region_changed.connect(self.aps_correlation.update)

        self.GACOSDialog = GACOSCorrectionDialog(model, spool)
        spool.actionGACOS.triggered.connect(self.GACOSDialog.show)

    @QtCore.pyqtSlot()
    def activateView(self):
        self.aps_correlation.activatePlot()
        self.main_widget.activatePlot()

    @QtCore.pyqtSlot()
    def deactivateView(self):
        self.aps_correlation.deactivatePlot()
        self.main_widget.activatePlot()

    def toggleAPS(self, checked):
        if checked:
            msg = QtWidgets.QMessageBox.question(
                self,
                "Apply APS to Scene",
                "Are you sure you want to apply APS to the scene?",
            )
            if msg == QtWidgets.QMessageBox.StandardButton.Yes:
                self.model.getScene().aps.set_enabled(True)
        else:
            self.model.getScene().aps.set_enabled(False)


class KiteAPSPlot(KitePlot):

    region_changed = QtCore.pyqtSignal()

    class TopoPatchROI(pg.RectROI):
        def _makePen(self):
            # Generate the pen color for this ROI based on its current state.
            if self.mouseHovering:
                return pen_roi_highlight
            else:
                return self.pen

    def __init__(self, model):
        self.components_available = {
            "displacement": ["Displacement", lambda sp: sp.scene.displacement]
        }
        self._component = "displacement"

        KitePlot.__init__(self, model=model, los_arrow=False, auto_downsample=True)

        llE, llN, sizeE, sizeN = self.model.aps.get_patch_coords()
        self.roi = self.TopoPatchROI(
            pos=(llE, llN), size=(sizeE, sizeN), sideScalers=True, pen=pen_roi
        )

        self.roi.setAcceptedMouseButtons(QtCore.Qt.RightButton)
        self.roi.sigRegionChangeFinished.connect(self.updateTopoRegion)

        self.addItem(self.roi)

    @QtCore.pyqtSlot()
    def updateTopoRegion(self):
        # data = self.roi.getArrayRegion(self.image.image, self.image)
        # data[data == 0.] = np.nan
        # if np.all(np.isnan(data)):
        #     return

        llE, llN = self.roi.pos()
        sizeE, sizeN = self.roi.size()
        patch_coords = (llE, llN, sizeE, sizeN)
        self.model.aps.set_patch_coords(*patch_coords)
        self.region_changed.emit()

    def activatePlot(self):
        self.enableHillshade()
        self.model.sigSceneChanged.connect(self.update)

    def deactivatePlot(self):
        self.model.sigSceneChanged.disconnect(self.update)


class KiteAPSCorrelation(KiteSubplot):

    MAXPOINTS = 10000

    def __init__(self, parent_plot):
        KiteSubplot.__init__(self, parent_plot)

        self.aps_correlation = pg.ScatterPlotItem(
            antialias=True, brush=brush_aps, pen=pen_aps, size=4
        )

        self.aps_model = pg.PlotDataItem(antialias=True, pen=pen_aps_model)

        self.legend = pg.LegendItem(offset=(0.0, 0.5))

        self.legend.setParentItem(self.plot.graphicsItem())
        self.legend.addItem(self.aps_model, "")

        self.addItem(self.aps_correlation)
        self.addItem(self.aps_model)

        self.plot.setLabels(bottom="Elevation (m)", left="Displacement (m)")

    @QtCore.pyqtSlot()
    def update(self):
        aps = self.model.aps
        elevation, displacement = aps.get_data()

        step = max(1, displacement.size // self.MAXPOINTS)

        self.aps_correlation.setData(elevation[::step], displacement[::step])

        slope, intercept = aps.get_correlation()
        elev = np.array([elevation.min(), elevation.max()])
        model = elev * slope + intercept

        self.aps_model.setData(elev, model)

        self.legend.items[-1][1].setText("Slope %.4f m / km" % (slope * km))

    def activatePlot(self):
        self.model.sigSceneChanged.connect(self.update)
        self.update()

    def deactivatePlot(self):
        self.model.sigSceneChanged.disconnect(self.update)


class GACOSCorrectionDialog(QtWidgets.QDialog):
    class GACOSPlot(KitePlot):
        def __init__(self, model, parent):
            self._component = "gacos_correction"
            self.parent = parent

            KitePlot.__init__(self, model)
            self.model = model

            self.update()

        def update(self):
            gacos = self.model.scene.gacos
            if not gacos.has_data():
                return

            corr = gacos.get_correction()
            self.image.updateImage(corr.T, autoLevels=True)
            self.transFromFrame()

    def __init__(self, model, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.model = model

        loadUi(get_resource("gacos_correction.ui"), baseinstance=self)
        self.closeButton.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogCloseButton)
        )

        self.gacos_plot = self.GACOSPlot(model, self)
        self.dockarea = dockarea.DockArea(self)

        self.dockarea.addDock(
            dockarea.Dock(
                "GACOS.get_correction()",
                widget=self.gacos_plot,
                size=(4, 4),
                autoOrientation=False,
            ),
            position="left",
        )

        self.horizontalLayoutPlot.addWidget(self.dockarea)
        self.loadGrids.released.connect(self.load_grids)
        self.toggleGACOS.released.connect(self.toggle_gacos)
        self.update_widgets()

    @QtCore.pyqtSlot()
    def load_grids(self):
        gacos = self.model.scene.gacos

        filenames, _ = QtWidgets.QFileDialog.getOpenFileNames(
            filter="GACOS file *.ztd (*.ztd)", caption="Load two GACOS APS Grids"
        )
        if not filenames:
            return
        if len(filenames) != 2:
            QtWidgets.QMessageBox.warning(
                self,
                "GACOS APS Correction Error",
                "We need two GACOS Grids to perform APS correction!",
            )
            return

        try:
            for filename in filenames:
                gacos.load(filename)
        except Exception as e:
            QtWidgets.QMessageBox.critical(self, "GACOS APS Correction Error", str(e))
            return
        self.gacos_plot.update()
        self.update_widgets()

    def update_widgets(self):
        gacos = self.model.scene.gacos

        tmpl_str = """
<b>Loaded GACOS Grids</b><br />
Grid 1 @{grd0.time}: {grd0.filename}<br />
Grid 2 @{grd1.time}: {grd1.filename}
"""

        if gacos.has_data():
            self.toggleGACOS.setEnabled(True)

            self.gridInformation.setText(
                tmpl_str.format(grd0=gacos.grids[0], grd1=gacos.grids[1])
            )

            if gacos.is_enabled():
                self.toggleGACOS.setText("Remove GACOS APS")
            else:
                self.toggleGACOS.setText("Apply GACOS APS")
        else:
            self.gridInformation.setText("No grids loaded.")
            self.toggleGACOS.setEnabled(False)

    def toggle_gacos(self):
        gacos = self.model.scene.gacos
        if gacos.is_enabled():
            gacos.set_enabled(False)
        else:
            gacos.set_enabled(True)
        self.update_widgets()


class EmpiricalAPSParams(KiteParameterGroup):
    def __init__(self, model, **kwargs):
        scene = model.getScene()
        kwargs["type"] = "group"
        kwargs["name"] = "Scene.APS (empirical)"

        KiteParameterGroup.__init__(self, model=model, model_attr="scene", **kwargs)

        p = {
            "name": "applied",
            "type": "bool",
            "value": scene.aps.config.applied,
            "tip": "detrend the scene",
        }
        self.applied = pTypes.SimpleParameter(**p)

        def toggle_applied(param, checked):
            self.model.getScene().aps.set_enabled(checked)

        self.applied.sigValueChanged.connect(toggle_applied)

        self.pushChild(self.applied)



#!/bin/python
import numpy as np
import scipy as sp

C = 299792458  # m/s


def squareMatrix(mat):
    if mat.shape[0] == mat.shape[1]:
        return mat
    min_a = np.argmin(mat.shape)
    max_a = np.argmax(mat.shape)

    width = mat.shape[max_a] - mat.shape[min_a]

    if min_a == 0:
        padding = ((width, 0), (0, 0))
    elif min_a == 1:
        padding = ((0, 0), (0, width))
    return np.pad(mat, pad_width=padding, mode="constant", constant_values=0.0)


def derampMatrix(displ):
    """Deramp through fitting a bilinear plane
    Data is also de-meaned
    """
    if displ.ndim != 2:
        raise TypeError("Displacement has to be 2-dim array")
    mx = np.nanmedian(displ, axis=0)
    my = np.nanmedian(displ, axis=1)

    ix = np.arange(mx.size)
    iy = np.arange(my.size)
    dx, cx, _, _, _ = sp.stats.linregress(ix[~np.isnan(mx)], mx[~np.isnan(mx)])
    dy, cy, _, _, _ = sp.stats.linregress(iy[~np.isnan(my)], my[~np.isnan(my)])

    rx = ix * dx
    ry = iy * dy
    data = displ - (rx[np.newaxis, :] + ry[:, np.newaxis])
    data -= np.nanmean(data)
    return data


def derampGMatrix(displ):
    """Deramp through lsq a bilinear plane
    Data is also de-meaned
    """
    if displ.ndim != 2:
        raise TypeError("Displacement has to be 2-dim array")

    # form a relative coordinate grid
    c_grid = np.mgrid[0 : displ.shape[0], 0 : displ.shape[1]]

    # separate and flatten coordinate grid into x and y vectors for each !point
    ix = c_grid[0].flat
    iy = c_grid[1].flat
    displ_f = displ.flat

    # reduce vectors taking out all NaN's
    displ_nonan = displ_f[np.isfinite(displ_f)]
    ix = ix[np.isfinite(displ_f)]
    iy = iy[np.isfinite(displ_f)]

    # form kernel/design derampMatrix (c, x, y)
    GT = np.matrix([np.ones(len(ix)), ix, iy])
    G = GT.T

    # generalized kernel matrix (quadtratic)
    GTG = GT * G
    # generalized inverse
    GTGinv = GTG.I

    # lsq estimates of ramp parameter
    ramp_paras = displ_nonan * (GTGinv * GT).T

    # ramp values
    ramp_nonan = ramp_paras * GT
    ramp_f = np.multiply(displ_f, 0.0)

    # insert ramp values in full vectors
    np.place(ramp_f, np.isfinite(displ_f), np.array(ramp_nonan).flatten())
    ramp_f = ramp_f.reshape(displ.shape[0], displ.shape[1])

    return displ - ramp_f


def trimMatrix(displ, data=None):
    """Trim displacement matrix from all NaN rows and columns"""
    if displ.ndim != 2:
        raise ValueError("Displacement has to be 2-dim array")

    if np.all(np.isnan(displ)):
        raise ValueError("Displacement is all NaN.")

    r1 = r2 = False
    c1 = c2 = False
    for r in range(displ.shape[0]):
        if not np.all(np.isnan(displ[r, :])):
            if r1 is False:
                r1 = r
            r2 = r
    for c in range(displ.shape[1]):
        if not np.all(np.isnan(displ[:, c])):
            if c1 is False:
                c1 = c
            c2 = c

    if data is not None:
        return data[r1 : (r2 + 1), c1 : (c2 + 1)]

    return displ[r1 : (r2 + 1), c1 : (c2 + 1)]


def greatCircleDistance(alat, alon, blat, blon):
    R1 = 6371009.0
    d2r = np.deg2rad
    sin = np.sin
    cos = np.cos
    a = (
        sin(d2r(alat - blat) / 2) ** 2
        + cos(d2r(alat)) * cos(d2r(blat)) * sin(d2r(alon - blon) / 2) ** 2
    )
    c = 2.0 * np.arctan2(np.sqrt(a), np.sqrt(1.0 - a))
    return R1 * c


def property_cached(func):
    var_name = "_cached_" + func.__name__
    func_doc = ":getter: *(Cached)*"
    if func.__doc__ is not None:
        func_doc += func.__doc__
    else:
        func_doc += " Undocumented"

    def cache_return(instance, *args, **kwargs):
        cache_return.__doc__ = func.__doc__
        if instance.__dict__.get(var_name, None) is None:
            instance.__dict__[var_name] = func(instance)
        return instance.__dict__[var_name]

    def cache_return_setter(instance, value):
        instance.__dict__[var_name] = value

    return property(fget=cache_return, fset=cache_return_setter, doc=func_doc)


def calcPrecision(data):
    # number of floating points:
    mn = np.nanmin(data)
    mx = np.nanmax(data)
    if not np.isfinite(mx) or np.isfinite(mn):
        return 3, 6
    precision = int(round(np.log10((100.0 / (mx - mn)))))
    if precision < 0:
        precision = 0
    # length of the number in the label:
    length = max(len(str(int(mn))), len(str(int(mx)))) + precision
    return precision, length


def formatScalar(v, ndigits=7):
    if np.isinf(v):
        return "inf"
    elif np.isnan(v):
        return "nan"

    if v % 1 == 0.0:
        return "{value:d}".format(value=v)

    if abs(v) < (10.0 ** -(ndigits - 2)):
        return "{value:e}".format(value=v)

    p = np.ceil(np.log10(np.abs(v)))
    if p <= 0.0:
        f = {"d": 1, "f": ndigits - 1}
    else:
        p = int(p)
        f = {"d": p, "f": ndigits - p}

    return "{value:{d}.{f}f}".format(value=v, **f)


class Subject(object):
    """
    Subject - Obsever model realization
    """

    def __init__(self):
        self._listeners = list()
        self._mute = False

    def __call__(self, *args, **kwargs):
        return self.notify(*args, **kwargs)

    def mute(self):
        self._mute = True

    def unmute(self):
        self._mute = False

    def subscribe(self, listener):
        """
        Subscribe a listening callback to this subject
        """
        if len(self._listeners) > 0:
            self._listeners.insert(0, listener)
        else:
            self._listeners.append(listener)

    def unsubscribe(self, listener):
        """
        Unsubscribe a listening callback from this subject
        """
        try:
            self._listeners.remove(listener)
        except Exception:
            raise AttributeError("%s was not subscribed!", listener.__name__)

    def unsubscribeAll(self):
        for l in self._listeners:
            self.unsubscribe(l)

    def notify(self, *args, **kwargs):
        if self._mute:
            return
        for l in self._listeners:
            if callable(l):
                self._call(l, *args, **kwargs)

    @staticmethod
    def _call(func, *args, **kwargs):
        try:
            for k in kwargs.keys():
                if k not in func.__code__.co_varnames:
                    k.pop(k)
        except AttributeError:
            kwargs = {}
        func(*args, **kwargs)


class ADict(dict):
    def __getattribute__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value


__all__ = """
Subject
""".split()



#!/usr/bin/env python3
import copy
import hashlib
import logging
import os.path as op
import time
from datetime import datetime as dt

import numpy as np
import utm
from pyrocko.dataset.topo import srtmgl3
from pyrocko.guts import Dict, Float, Object, String, StringChoice, Timestamp, load
from pyrocko.orthodrome import latlon_to_ne, latlon_to_ne_numpy, ne_to_latlon  # noqa
from scipy import interpolate

from kite import scene_io
from kite.aps import APS, APSConfig
from kite.covariance import CovarianceConfig
from kite.deramp import Deramp, DerampConfig
from kite.gacos import GACOSConfig, GACOSCorrection
from kite.quadtree import QuadtreeConfig
from kite.scene_mask import PolygonMask, PolygonMaskConfig
from kite.util import Subject, property_cached


def read(filename):
    try:
        return Scene.load(filename)
    except (ImportError, UserIOWarning):
        pass
    try:
        scene = Scene()
        scene.import_data(filename)
        return scene
    except ImportError:
        pass
    raise ImportError("Could not read file %s" % filename)


def _setDataNumpy(obj, variable, value):
    if isinstance(value, np.ndarray):
        return obj.__setattr__(variable, value)
    else:
        raise TypeError("value must be of type numpy.ndarray")


class UserIOWarning(UserWarning):
    pass


class SceneError(Exception):
    pass


class FrameConfig(Object):
    """Config object holding :class:`kite.scene.Scene` configuration"""

    llLat = Float.T(default=0.0, help="Scene latitude of lower left corner")
    llLon = Float.T(default=0.0, help="Scene longitude of lower left corner")
    dN = Float.T(default=25.0, help="Scene pixel spacing in north, give [m] or [deg]")
    dE = Float.T(default=25.0, help="Scene pixel spacing in east, give [m] or [deg]")
    spacing = StringChoice.T(
        choices=("degree", "meter"), default="meter", help="Unit of pixel space"
    )

    def __init__(self, *args, **kwargs):
        self.old_import = False
        mapping = {"dE": "dLon", "dN": "dLat"}

        for new, old in mapping.items():
            if old in kwargs:
                kwargs[new] = kwargs.pop(old)
                kwargs["spacing"] = "degree"
                self.old_import = True

        Object.__init__(self, *args, **kwargs)


class Frame(object):
    """Frame holding geographical references for :class:`kite.scene.Scene`

    The pixel spacing is given by ``dE`` and ``dN`` which can meters or degree.
    """

    def __init__(self, scene, config=None):
        self.evChanged = Subject()
        self._scene = scene
        self._log = scene._log.getChild("Frame")

        self.N = None
        self.E = None

        self.llEutm = None
        self.llNutm = None
        self.utm_zone = None
        self.utm_zone_letter = None
        self._meter_grid = None

        self._updateConfig(config or FrameConfig())
        self._scene.evConfigChanged.subscribe(self._updateConfig)
        self._scene.evChanged.subscribe(self.updateExtent)

    def _updateConfig(self, config=None):
        if config is not None:
            self.config = config
        elif self.config != self._scene.config.frame:
            self.config = self._scene.config.frame
        else:
            return

        if self.config.old_import:
            self._log.warning("Importing an old kite format...")
            self._log.warning("Please check your pixel spacing - dE, dN!")
        self.updateExtent()

    def updateExtent(self):
        if self._scene.cols == 0 or self._scene.rows == 0:
            return

        self.cols = self._scene.cols
        self.rows = self._scene.rows

        (
            self.llEutm,
            self.llNutm,
            self.utm_zone,
            self.utm_zone_letter,
        ) = utm.from_latlon(self.llLat, self.llLon)

        self.E = None
        self.N = None

        self.gridE = None
        self.gridN = None
        self._meter_grid = None
        self.coordinates = None

        self.config.regularize()
        self.evChanged.notify()

    @property
    def llLat(self):
        return self.config.llLat

    @llLat.setter
    def llLat(self, llLat):
        self.config.llLat = llLat
        self.updateExtent()

    @property
    def llLon(self):
        return self.config.llLon

    @llLon.setter
    def llLon(self, llLon):
        self.config.llLon = llLon
        self.updateExtent()

    @property
    def dN(self):
        return self.config.dN

    @dN.setter
    def dN(self, dN):
        self.config.dN = dN
        self.updateExtent()

    @property
    def dE(self):
        return self.config.dE

    @dE.setter
    def dE(self, dE):
        self.config.dE = dE
        self.updateExtent()

    @property
    def dEmeter(self):
        if self.isMeter():
            return self.dE

        _, dEmeter = latlon_to_ne(
            self.llLat, self.llLon, self.llLat, self.llLon + self.dE * self.cols
        )
        return dEmeter / self.cols

    @property
    def dNmeter(self):
        if self.isMeter():
            return self.dN
        dNmeter, _ = latlon_to_ne(
            self.llLat, self.llLon, self.llLat + self.dN * self.rows, self.llLon
        )
        return dNmeter / self.rows

    @property
    def dEdegree(self):
        if self.isDegree():
            return self.dE

        lat, lon = ne_to_latlon(self.llLat, self.llLon, 0.0, self.dE * self.cols)
        distLon = lon - self.llLon
        return distLon / self.cols

    @property
    def dNdegree(self):
        if self.isDegree():
            return self.dE

        lat, lon = ne_to_latlon(self.llLat, self.llLon, self.dN * self.rows, 0.0)
        distLat = lat - self.llLat
        return distLat / self.rows

    @property
    def spacing(self):
        return self.config.spacing

    @spacing.setter
    def spacing(self, unit):
        self.config.spacing = unit

    @property_cached
    def E(self):
        return np.arange(self.cols) * self.dE

    @property_cached
    def Emeter(self):
        return np.arange(self.cols) * self.dEmeter

    @property_cached
    def N(self):
        return np.arange(self.rows) * self.dN

    @property
    def lengthE(self):
        return self.cols * self.dE

    @property
    def lengthN(self):
        return self.rows * self.dN

    @property_cached
    def Nmeter(self):
        return np.arange(self.rows) * self.dNmeter

    @property_cached
    def gridE(self):
        """Grid holding local east coordinates of all pixels in ``NxM`` matrix
            of :attr:`~kite.Scene.displacement`.

        :type: :class:`numpy.ndarray`, size ``NxM``
        """
        valid_data = np.isnan(self._scene.displacement)
        gridE = np.repeat(self.E[np.newaxis, :], self.rows, axis=0)
        return np.ma.masked_array(gridE, valid_data, fill_value=np.nan)

    @property_cached
    def gridN(self):
        """Grid holding local north coordinates of all pixels in ``NxM`` matrix
            of :attr:`~kite.Scene.displacement`.

        :type: :class:`numpy.ndarray`, size ``NxM``
        """
        valid_data = np.isnan(self._scene.displacement)
        gridN = np.repeat(self.N[:, np.newaxis], self.cols, axis=1)
        return np.ma.masked_array(gridN, valid_data, fill_value=np.nan)

    def _calculateMeterGrid(self):
        if self.isMeter():
            raise ValueError(
                "Frame is defined in meter! " "Use gridE and gridN for meter grids"
            )

        if self._meter_grid is None:
            self._log.debug("Transforming latlon grid to meters...")
            gridN, gridE = latlon_to_ne_numpy(
                self.llLat,
                self.llLon,
                self.llLat + self.gridN.data.ravel(),
                self.llLon + self.gridE.data.ravel(),
            )

            valid_data = np.isnan(self._scene.displacement)
            gridE = np.ma.masked_array(
                gridE.reshape(self.gridE.shape), valid_data, fill_value=np.nan
            )
            gridN = np.ma.masked_array(
                gridN.reshape(self.gridN.shape), valid_data, fill_value=np.nan
            )
            self._meter_grid = (gridE, gridN)

        return self._meter_grid

    @property_cached
    def gridEmeter(self):
        if self.isMeter():
            return self.gridE

        return self._calculateMeterGrid()[0]

    @property_cached
    def gridNmeter(self):
        if self.isMeter():
            return self.gridN
        return self._calculateMeterGrid()[1]

    @property_cached
    def coordinates(self):
        """Local east and north coordinates of all pixels in
           ``Nx2`` matrix.

        :type: :class:`numpy.ndarray`, size ``Nx2``
        """
        coords = np.empty((self.rows * self.cols, 2))
        coords[:, 0] = np.repeat(self.E[np.newaxis, :], self.rows, axis=0).flatten()
        coords[:, 1] = np.repeat(self.N[:, np.newaxis], self.cols, axis=1).flatten()

        if self.isMeter():
            coords = ne_to_latlon(self.llLat, self.llLon, *coords.T)
            coords = np.array(coords).T

        else:
            coords[:, 0] += self.llLon
            coords[:, 1] += self.llLat

        return coords

    @property_cached
    def coordinatesMeter(self):
        """Local east and north coordinates [m] of all pixels in
           ``NxM`` matrix.

        :type: :class:`numpy.ndarray`, size ``NxM``
        """
        coords = np.empty((self.rows * self.cols, 2))
        coords[:, 0] = np.repeat(
            self.Emeter[np.newaxis, :], self.rows, axis=0
        ).flatten()
        coords[:, 1] = np.repeat(
            self.Nmeter[:, np.newaxis], self.cols, axis=1
        ).flatten()
        return coords

    def mapENMatrix(self, E, N):
        """Local map coordinates in east and north to matrix
            row and column

        :param E: Easting in local coordinates
        :type E: float
        :param N: Northing in local coordinates
        :type N: float
        :returns: Row and column
        :rtype: tuple (int), (row, column)
        """
        row = round(E / self.dE) if E > 0 else 0
        col = round(N / self.dN) if N > 0 else 0
        return int(row), int(col)

    @property
    def shape(self):
        return self._scene.shape

    def isMeter(self):
        return self.config.spacing == "meter"

    def isDegree(self):
        return self.config.spacing == "degree"

    @property
    def npixel(self):
        return self.cols * self.rows

    def __eq__(self, other):
        return (
            self.llLat == other.llLat
            and self.llLon == other.llLon
            and self.dE == other.dE
            and self.dN == other.dN
            and self.rows == other.rows
            and self.cols == other.cols
        )


class Meta(Object):
    """Meta configuration for ``Scene``."""

    scene_title = String.T(default="Unnamed Scene", help="Scene title")
    scene_id = String.T(default="None", help="Scene identification")
    satellite_name = String.T(
        default="Undefined Mission", help="Satellite mission name"
    )
    wavelength = Float.T(optional=True, help="Wavelength in [m]")
    orbital_node = StringChoice.T(
        choices=["Ascending", "Descending", "Undefined"],
        default="Undefined",
        help="Orbital direction, ascending/descending",
    )
    time_master = Timestamp.T(
        default=1481116161.930574, help="Timestamp for master acquisition"
    )
    time_slave = Timestamp.T(
        default=1482239325.482, help="Timestamp for slave acquisition"
    )
    extra = Dict.T(default={}, help="Extra header information")
    filename = String.T(optional=True)

    def __init__(self, *args, **kwargs):
        self.old_import = False

        mapping = {"orbit_direction": "orbital_node"}

        for old, new in mapping.items():
            if old in kwargs.keys():
                kwargs[new] = kwargs.pop(old, None)
                self.old_import = True

        Object.__init__(self, *args, **kwargs)

    @property
    def time_separation(self):
        """
        :getter: Absolute time difference between ``time_master``
                 and ``time_slave``
        :type: timedelta
        """
        return dt.fromtimestamp(self.time_slave) - dt.fromtimestamp(self.time_master)


class SceneConfig(Object):
    """Configuration object, gathering ``kite.Scene`` and
    sub-objects configuration.
    """

    meta = Meta.T(default=Meta.D(), help="Scene metainformation")
    frame = FrameConfig.T(default=FrameConfig.D(), help="Frame/reference configuration")
    quadtree = QuadtreeConfig.T(default=QuadtreeConfig.D(), help="Quadtree parameters")
    covariance = CovarianceConfig.T(
        default=CovarianceConfig.D(), help="Covariance parameters"
    )
    aps = APSConfig.T(default=APSConfig.D(), help="Empirical APS correction")
    gacos = GACOSConfig.T(default=GACOSConfig.D(), help="GACOS APS correction")
    polygon_mask = PolygonMaskConfig.T(
        default=PolygonMaskConfig.D(), help="Displacement mask polygon"
    )
    deramp = DerampConfig.T(default=DerampConfig.D(), help="Displacement deramp config")

    @property
    def old_import(self):
        return self.frame.old_import


def dynamicmethod(func):
    """Decorator for dynamic classmethod / instancemethod declaration"""

    def dynclassmethod(*args, **kwargs):
        if isinstance(args[0], Scene):
            return func(*args, **kwargs)
        else:
            return func(Scene(), *args, **kwargs)

    dynclassmethod.__doc__ = func.__doc__
    dynclassmethod.__name__ = func.__name__
    return dynclassmethod


class BaseScene(object):
    def __init__(self, **kwargs):
        self._log = logging.getLogger(self.__class__.__name__)

        self.evChanged = Subject()
        self.evConfigChanged = Subject()

        self._displacement = None
        self._displacement_px_var = None
        self._phi = None
        self._theta = None
        self._los_factors = None
        self.cols = 0
        self.rows = 0
        self.los = LOSUnitVectors(scene=self)

        self._elevation = {}

        frame_config = kwargs.pop("frame_config", FrameConfig())

        for fattr in ("llLat", "llLon", "dLat", "dLon"):
            coord = kwargs.pop(fattr, None)
            if coord is not None:
                frame_config.__setattr__(fattr, coord)
        self.frame = Frame(scene=self, config=frame_config)

        for attr in ("displacement", "displacement_px_var", "theta", "phi"):
            data = kwargs.pop(attr, None)
            if data is not None:
                self.__setattr__(attr, data)

    @property
    def displacement(self):
        """Surface displacement in meter on a regular grid.

        :setter: Set the unwrapped InSAR displacement.
        :getter: Return the displacement matrix.
        :type: :class:`numpy.ndarray`, ``NxM``
        """
        return self._displacement

    @displacement.setter
    def displacement(self, value):
        _setDataNumpy(self, "_displacement", value)
        self.rows, self.cols = self._displacement.shape
        self.evChanged.notify()

    @property
    def displacement_px_var(self):
        """Variance of the surface displacement per pixel.
        Same dimension as displacement.

        :setter: Set standard deviation of of the displacement.
        :getter: Return the standard deviation matrix.
        :type: :class:`numpy.ndarray`, ``NxM``
        """
        return self._displacement_px_var

    @displacement_px_var.setter
    def displacement_px_var(self, value):
        self._displacement_px_var = value

    @property
    def displacement_mask(self):
        """Displacement :attr:`numpy.nan` mask

        :type: :class:`numpy.ndarray`, dtype :class:`numpy.bool`
        """
        return ~np.isfinite(self.displacement)

    @property
    def shape(self):
        return self.displacement.shape

    @property
    def phi(self):
        """Horizontal angle towards satellite :abbr:`line of sight (LOS)`
        in radians counter-clockwise from East.

        .. important ::

            Kite's convention is:

            * :math:`0` is **East**
            * :math:`\\frac{\\pi}{2}` is **North**!

        :setter: Set the phi matrix for scene's displacement, can be ``int``
                 for static look vector.
        :type: :class:`numpy.ndarray`, size same as
               :attr:`~kite.Scene.displacement` or int
        """
        return self._phi

    @phi.setter
    def phi(self, value):
        if isinstance(value, float):
            self._phi = value
        else:
            _setDataNumpy(self, "_phi", value)
        self.phiDeg = None
        self.los_rotation_factors = None
        self.evChanged.notify()

    @property
    def theta(self):
        """Theta is the look vector elevation angle towards satellite from
        the horizon in radians. Matrix of theta towards satellite's
        :abbr:`line of sight (LOS)`.

        .. important ::

            Kite convention!

            * :math:`-\\frac{\\pi}{2}` is **Down**
            * :math:`\\frac{\\pi}{2}` is **Up**

        :setter: Set the theta matrix for scene's displacement, can be ``int``
                 for static look vector.
        :type: :class:`numpy.ndarray`, size same as
               :attr:`~kite.Scene.displacement` or int
        """
        return self._theta

    @theta.setter
    def theta(self, value):
        if isinstance(value, float):
            self._theta = value
        else:
            _setDataNumpy(self, "_theta", value)
        self.thetaDeg = None
        self.los_rotation_factors = None
        self.evChanged.notify()

    @property_cached
    def thetaDeg(self):
        """LOS elevation angle in degree, ``NxM`` matrix like
            :class:`kite.Scene.theta`

        :type: :class:`numpy.ndarray`
        """
        return np.rad2deg(self.theta)

    @property_cached
    def phiDeg(self):
        """LOS horizontal orientation angle in degree,
            counter-clockwise from East,``NxM`` matrix like
            :class:`kite.Scene.phi`

        :type: :class:`numpy.ndarray`
        """
        return np.rad2deg(self.phi)

    @property_cached
    def los_rotation_factors(self):
        """ Trigonometric factors to rotate displacement matrices towards LOS

        Rotation is as follows:

        ..
            displacement_los =\
                (los_rotation_factors[:, :, 0] * -down +
                 los_rotation_factors[:, :, 1] * east +
                 los_rotation_factors[:, :, 2] * north)

        :returns: Factors for rotation
        :rtype: :class:`numpy.ndarray`, ``NxMx3``
        :raises: AttributeError
        """
        if self.theta.size != self.phi.size:
            raise AttributeError(
                "LOS angles inconsistent with provided" " coordinate shape."
            )
        if self._los_factors is None:
            self._los_factors = np.empty((self.theta.shape[0], self.theta.shape[1], 3))
            self._los_factors[:, :, 0] = np.sin(self.theta)
            self._los_factors[:, :, 1] = np.cos(self.theta) * np.cos(self.phi)
            self._los_factors[:, :, 2] = np.cos(self.theta) * np.sin(self.phi)
        return self._los_factors

    def get_elevation(self, interpolation="nearest_neighbor"):
        assert interpolation in ("nearest_neighbor", "bivariate")

        if self._elevation.get(interpolation, None) is None:
            self._log.debug("Getting elevation...")
            # region = llLon, urLon, llLat, urLon
            coords = self.frame.coordinates
            lons = coords[:, 0]
            lats = coords[:, 1]

            region = (lons.min(), lons.max(), lats.min(), lats.max())
            if not srtmgl3.covers(region):
                raise AssertionError("Region is outside of SRTMGL3 topo dataset")

            tile = srtmgl3.get(region)
            if not tile:
                raise AssertionError("Cannot get SRTMGL3 topo dataset")

            if interpolation == "nearest_neighbor":
                iy = np.rint((lats - tile.ymin) / tile.dy).astype(np.intp)
                ix = np.rint((lons - tile.xmin) / tile.dx).astype(np.intp)

                elevation = tile.data[(iy, ix)]

            elif interpolation == "bivariate":
                interp = interpolate.RectBivariateSpline(tile.y(), tile.x(), tile.data)
                elevation = interp(lats, lons, grid=False)

            elevation = elevation.reshape(self.rows, self.cols)
            self._elevation[interpolation] = elevation

        return self._elevation[interpolation]

    def __neg__(self):
        ret = copy.deepcopy(self)
        ret.displacement *= -1
        return ret

    def __add__(self, other, copy_obj=True):
        if copy_obj:
            ret = copy.deepcopy(self)
        else:
            ret = self

        if not ret.frame == other.frame:
            raise ValueError("Scene frames do not align!")
        ret.displacement += other.displacement

        tmin = (
            ret.meta.time_master
            if ret.meta.time_master < other.meta.time_master
            else other.meta.time_master
        )

        tmax = (
            ret.meta.time_slave
            if ret.meta.time_slave > other.meta.time_slave
            else other.meta.time_slave
        )

        ret.meta.time_master = tmin
        ret.meta.time_slave = tmax
        return ret

    def __sub__(self, other):
        return self.__add__(-other)

    def __isub__(self, scene):
        return self.__add__(-scene, copy_obj=False)

    def __iadd__(self, scene):
        return self.__add__(scene, copy_obj=False)


class Scene(BaseScene):
    """Scene of unwrapped InSAR ground displacements measurements

    :param config: Configuration object
    :type config: :class:`~kite.scene.SceneConfig`, optional

    Optional parameters

    :param displacement: Displacement in [m]
    :type displacement: :class:`numpy.ndarray`, NxM, optional
    :param theta: Theta look angle, see :attr:`BaseScene.theta`
    :type theta: :class:`numpy.ndarray`, NxM, optional
    :param phi: Phi look angle, see :attr:`BaseScene.phi`
    :type phi: :class:`numpy.ndarray`, NxM, optional

    :param llLat: Lower left latitude in [deg]
    :type llLat: float, optional
    :param llLon: Lower left longitude in [deg]
    :type llLon: float, optional
    :param dLat: Pixel spacing in latitude [deg or m]
    :type dLat: float, optional
    :param dLon: Pixel spacing in longitude [deg or m]
    :type dLon: float, optional
    """

    def __init__(self, config=None, **kwargs):
        self.config = config or SceneConfig()
        self.meta = self.config.meta

        BaseScene.__init__(self, frame_config=self.config.frame, **kwargs)

        # wiring special methods
        self.import_data = self._import_data

        self.aps = APS(self, self.config.aps)
        self.gacos = GACOSCorrection(self, self.config.gacos)
        self.polygon_mask = PolygonMask(self, self.config.polygon_mask)
        self.deramp = Deramp(self, self.config.deramp)

        self._proc_displacement = None

        self.processing_states = {
            self.gacos: None,
            self.aps: None,
            self.polygon_mask: None,
            self.deramp: None,
        }

    @property
    def displacement(self):
        if self.has_processing_changed() or self._proc_displacement is None:

            self._proc_displacement = self._displacement.copy()
            for plugin, state in self.processing_states.items():
                self.processing_states[plugin] = plugin.get_state_hash()
                if not plugin.is_enabled():
                    continue

                t = time.time()
                plugin.apply(self._proc_displacement)
                self._log.debug(
                    "applied %s in %.4f s", plugin.__class__.__name__, time.time() - t
                )

        return self._proc_displacement

    @displacement.setter
    def displacement(self, value):
        _setDataNumpy(self, "_displacement", value)
        self.rows, self.cols = self._displacement.shape
        self.evChanged.notify()

    @property_cached
    def quadtree(self):
        """Instantiates the scene's quadtree.

        :type: :class:`kite.quadtree.Quadtree`
        """
        self._log.debug("Creating kite.Quadtree instance")
        from kite.quadtree import Quadtree

        return Quadtree(scene=self, config=self.config.quadtree)

    @property_cached
    def covariance(self):
        """Instantiates the scene's covariance attribute.

        :type: :class:`kite.covariance.Covariance`
        """
        self._log.debug("Creating kite.Covariance instance")
        from kite.covariance import Covariance

        return Covariance(scene=self, config=self.config.covariance)

    @property_cached
    def plot(self):
        """Shows a simple plot of the scene's displacement"""
        self._log.debug("Creating kite.ScenePlot instance")
        from kite.plot2d import ScenePlot

        return ScenePlot(self)

    def has_processing_changed(self):
        for plugin, state in self.processing_states.items():
            if state != plugin.get_state_hash():
                self._log.debug(
                    "processing states changed: %s", plugin.__class__.__name__
                )
                return True
        return False

    def get_plugin_state_hash(self):
        sha = hashlib.sha1()
        for plugin, state in self.processing_states.items():
            sha.update(plugin.get_state_hash().encode())

        return sha.hexdigest()

    def get_state_hash(self):
        sha = hashlib.sha1()
        for plugin, state in self.processing_states.items():
            sha.update(plugin.get_state_hash().encode())

        sha.update(self.covariance.get_state_hash().encode())
        sha.update(self.quadtree.get_state_hash().encode())
        return sha.hexdigest()

    def spool(self):
        """Start the spool user interface :class:`~kite.spool.Spool` to inspect
        the scene.
        """
        if self.displacement is None:
            raise SceneError("Can not display an empty scene.")

        from kite.spool import spool

        spool(scene=self)

    def _testImport(self):
        try:
            self.frame.E
            self.frame.N
            self.frame.gridE
            self.frame.gridN
            self.frame.dE
            self.frame.dN
            self.displacement
            self.theta
            self.phi
        except Exception as e:
            self._log.exception(e)
            raise ImportError("Something went wrong during import - " "see Exception!")

    def save(self, filename=None):
        """Save kite scene to kite file structure

        Saves the current scene meta information and UTM frame to a YAML
        (``.yml``) file. Numerical data (:attr:`~kite.Scene.displacement`,
        :attr:`~kite.Scene.theta` and :attr:`~kite.Scene.phi`)
        are saved as binary files from :class:`numpy.ndarray`.

        :param filename: Filenames to save scene to, defaults to
            ' :attr:`~kite.Scene.meta.scene_id` ``_``
            :attr:`~kite.Scene.meta.scene_view`
        :type filename: str, optional
        """
        filename = filename or "%s_%s" % (self.meta.scene_id, self.meta.scene_view)
        _file, ext = op.splitext(filename)
        filename = _file if ext in [".yml", ".npz"] else filename

        components = ["_displacement", "theta", "phi"]
        self._log.debug("Saving scene data to %s.npz" % filename)

        np.savez("%s.npz" % (filename), *[getattr(self, arr) for arr in components])

        self.gacos.save(op.dirname(op.abspath(filename)))

        self.saveConfig("%s.yml" % filename)

    def saveConfig(self, filename):
        _file, ext = op.splitext(filename)
        filename = filename if ext in [".yml"] else filename + ".yml"
        self._log.debug("Saving scene config to %s" % filename)
        self.config.regularize()
        self.config.dump(filename="%s" % filename, header="kite.Scene YAML Config")

    @classmethod
    def load(cls, filename):
        """Load a kite scene from file ``filename.[npz,yml]``
        structure.

        :param filename: Filenames the scene data is saved under
        :type filename: str
        :returns: Scene object from data resources
        :rtype: :class:`~kite.Scene`
        """
        filename = op.abspath(filename)
        basename = op.splitext(filename)[0]

        try:
            data = np.load("%s.npz" % basename)
            displacement = data["arr_0"]
            theta = data["arr_1"]
            phi = data["arr_2"]
        except IOError:
            raise UserIOWarning("Could not load data from %s.npz" % basename)

        try:
            config = load(filename="%s.yml" % basename)
            config.meta.filename = filename
        except IOError:
            raise UserIOWarning("Could not load %s.yml" % basename)

        scene = cls(displacement=displacement, theta=theta, phi=phi, config=config)
        scene._log.debug("Loading from %s[.npz,.yml]", basename)

        scene.meta.filename = filename

        scene._testImport()
        return scene

    def load_config(self, filename):
        self._log.debug("Loading config from %s", filename)
        self.config = load(filename=filename)
        self.meta = self.config.meta

        self.evConfigChanged.notify()

    @dynamicmethod
    def _import_data(self, path, **kwargs):
        """Import displacement data from foreign file format.

        :param path: Filename of resource to import
        :type path: str
        :param kwargs: keyword arguments passed to import function
        :type kwargs: dict
        :returns: Scene from path
        :rtype: :class:`~kite.Scene`
        :raises: TypeError
        """
        scene = self
        if not op.isfile(path) and not op.isdir(path):
            raise ImportError("File %s does not exist!" % path)

        data = None

        for mod_name in scene_io.__all__:
            cls = getattr(__import__("kite.scene_io", fromlist=mod_name), mod_name)
            module = cls()
            if module.validate(path, **kwargs):
                scene._log.debug("Importing %s using %s module" % (path, mod_name))
                data = module.read(path, **kwargs)
                break
        if data is None:
            raise ImportError("Could not recognize format for %s" % path)

        scene.meta.filename = op.abspath(path)
        return scene._import_from_dict(scene, data)

    _class_list = map("* :class:`~kite.scene_io.{}`".format, scene_io.__all__)
    _import_data.__doc__ += (
        "\nSupported import for unwrapped InSAR data are:\n\n{}\n".format(
            "\n".join(_class_list)
        )
    )
    for mod_name in scene_io.__all__:
        cls = getattr(__import__("kite.scene_io", fromlist=mod_name), mod_name)

        _import_data.__doc__ += "\n**{name}**\n\n{doc}".format(
            name=mod_name, doc=cls.__doc__
        )
    import_data = staticmethod(_import_data)

    @staticmethod
    def _import_from_dict(scene, data):
        for sk in ["theta", "phi", "displacement"]:
            setattr(scene, sk, data[sk])

        for fk, fv in data["frame"].items():
            setattr(scene.frame, fk, fv)

        for mk, mv in data["meta"].items():
            if mv is not None:
                setattr(scene.meta, mk, mv)
        scene.meta.extra.update(data["extra"])
        scene.frame.updateExtent()

        scene._testImport()
        return scene

    def __str__(self):
        return self.config.__str__()


class LOSUnitVectors(object):
    """Decompose line-of-sight (LOS) angles derived from
    :attr:`~kite.Scene.displacement` to unit vector.
    """

    def __init__(self, scene):
        self._scene = scene
        self._scene.evChanged.subscribe(self._flush_vectors)

    def _flush_vectors(self):
        self.unitE = None
        self.unitN = None
        self.unitU = None

    @property_cached
    def unitE(self):
        """Unit vector east component, ``NxM`` matrix like
            :attr:`~kite.Scene.displacement`
        :type: :class:`numpy.ndarray`
        """
        return self._scene.los_rotation_factors[:, :, 1]

    @property_cached
    def unitN(self):
        """Unit vector north component, ``NxM`` matrix like
            :attr:`~kite.Scene.displacement`
        :type: :class:`numpy.ndarray`
        """
        return self._scene.los_rotation_factors[:, :, 2]

    @property_cached
    def unitU(self):
        """Unit vector vertical (up) component, ``NxM`` matrix like
            :attr:`~kite.Scene.displacement`
        :type: :class:`numpy.ndarray`
        """
        return self._scene.los_rotation_factors[:, :, 0]


class TestScene(Scene):
    """Test scenes for synthetic displacement"""

    @classmethod
    def createGauss(cls, nx=512, ny=512, noise=None, **kwargs):
        scene = cls()
        scene.meta.scene_title = "Synthetic Displacement | Gaussian"
        scene = cls._prepareSceneTest(scene, nx, ny)

        scene.displacement = scene._gaussAnomaly(scene.frame.E, scene.frame.N, **kwargs)
        if noise is not None:
            cls.addNoise(noise)
        return scene

    @classmethod
    def createRandom(cls, nx=512, ny=512, **kwargs):
        scene = cls()
        scene.meta.title = "Synthetic Displacement | Uniform Random"
        scene = cls._prepareSceneTest(scene, nx, ny)

        rand_state = np.random.RandomState(seed=kwargs.pop("seed", None))
        scene.displacement = (rand_state.rand(nx, ny) - 0.5) * 2

        return scene

    @classmethod
    def createSine(
        cls, nx=512, ny=512, kE=0.0041, kN=0.0061, amplitude=1.0, noise=0.5, **kwargs
    ):
        scene = cls()
        scene.meta.title = "Synthetic Displacement | Sine"
        scene = cls._prepareSceneTest(scene, nx, ny)

        E, N = np.meshgrid(scene.frame.E, scene.frame.N)
        displ = np.zeros_like(E)

        kE = np.random.rand(3) * kE
        kN = np.random.rand(3) * kN

        for ke in kE:
            phase = np.random.randn(1)[0]
            displ += np.sin(ke * E + phase)
        for kn in kN:
            phase = np.random.randn(1)[0]
            displ += np.sin(kn * N + phase)
        displ -= np.mean(displ)

        scene.displacement = displ * amplitude
        if noise is not None:
            scene.addNoise(noise)
        return scene

    @classmethod
    def createFractal(
        cls,
        nE=1024,
        nN=1024,
        beta=[5.0 / 3, 8.0 / 3, 2.0 / 3],
        regime=[0.15, 0.99, 1.0],
        amplitude=1.0,
    ):
        scene = cls()
        scene.meta.title = "Synthetic Displacement | Fractal Noise (Hanssen, 2001)"
        scene = cls._prepareSceneTest(scene, nE, nN)
        if (nE + nN) % 2 != 0:
            raise ArithmeticError("Dimensions of synthetic scene must " "both be even!")

        dE, dN = (scene.frame.dE, scene.frame.dN)

        rfield = np.random.rand(nE, nN)
        spec = np.fft.fft2(rfield)

        kE = np.fft.fftfreq(nE, dE)
        kN = np.fft.fftfreq(nN, dN)
        k_rad = np.sqrt(kN[:, np.newaxis] ** 2 + kE[np.newaxis, :] ** 2)

        regime = np.array(regime)
        k0 = 0.0
        k1 = regime[0] * k_rad.max()
        k2 = regime[1] * k_rad.max()

        r0 = np.logical_and(k_rad > k0, k_rad < k1)
        r1 = np.logical_and(k_rad >= k1, k_rad < k2)
        r2 = k_rad >= k2

        beta = np.array(beta)
        # From Hanssen (2001)
        #   beta+1 is used as beta, since, the power exponent
        #   is defined for a 1D slice of the 2D spectrum:
        #   austin94: "Adler, 1981, shows that the surface profile
        #   created by the intersection of a plane and a
        #   2-D fractal surface is itself fractal with
        #   a fractal dimension  equal to that of the 2D
        #   surface decreased by one."
        beta += 1.0
        # From Hanssen (2001)
        #   The power beta/2 is used because the power spectral
        #   density is proportional to the amplitude squared
        #   Here we work with the amplitude, instead of the power
        #   so we should take sqrt( k.^beta) = k.^(beta/2)  RH
        # beta /= 2.

        amp = np.zeros_like(k_rad)
        amp[r0] = k_rad[r0] ** -beta[0]
        amp[r0] /= amp[r0].max()

        amp[r1] = k_rad[r1] ** -beta[1]
        amp[r1] /= amp[r1].max() / amp[r0].min()

        amp[r2] = k_rad[r2] ** -beta[2]
        amp[r2] /= amp[r2].max() / amp[r1].min()

        amp[k_rad == 0.0] = amp.max()

        spec *= amplitude * np.sqrt(amp)
        disp = np.abs(np.fft.ifft2(spec))
        disp -= np.mean(disp)

        scene.displacement = disp
        return scene

    def addNoise(self, noise_amplitude=1.0, seed=None):
        rand = np.random.RandomState(seed)
        noise = rand.randn(*self.displacement.shape) * noise_amplitude
        self.displacement += noise

    @staticmethod
    def _prepareSceneTest(scene, nE=512, nN=512):
        scene.frame.llLat = 0.0
        scene.frame.llLon = 0.0
        scene.frame.dLat = 5e-4
        scene.frame.dLon = 5e-4
        # scene.frame.E = np.arange(nE) * 50.
        # scene.frame.N = np.arange(nN) * 50.
        scene.theta = np.repeat(np.linspace(0.8, 0.85, nE), nN).reshape((nE, nN))
        scene.phi = np.rot90(scene.theta)
        scene.displacement = np.zeros((nE, nN))
        return scene

    @staticmethod
    def _gaussAnomaly(
        x, y, sigma_x=0.007, sigma_y=0.005, amplitude=3.0, x0=None, y0=None
    ):
        if x0 is None:
            x0 = x.min() + abs(x.max() - x.min()) / 2
        if y0 is None:
            y0 = y.min() + abs(y.max() - y.min()) / 2
        X, Y = np.meshgrid(x, y)

        gauss_anomaly = amplitude * np.exp(
            -(((X - x0) ** 2 / 2 * sigma_x**2) + (Y - y0) ** 2 / 2 * sigma_y**2)
        )

        return gauss_anomaly


__all__ = ["Scene", "SceneConfig"]


if __name__ == "__main__":
    testScene = TestScene.createGauss()



import numpy as np

from kite.sources import EllipsoidSource, PointCompoundSource

from .base import PointSourceROI, SourceDelegate, SourceEditDialog

d2r = np.pi / 180.0
r2d = 180.0 / np.pi


class EllipsoidSourceDelegate(SourceDelegate):

    __represents__ = "EllipsoidSource"

    display_backend = "Compound Model"
    display_name = "EllipsoidSource"

    parameters = [
        "easting",
        "northing",
        "depth",
        "length_x",
        "length_y",
        "length_z",
        "rotation_x",
        "rotation_y",
        "rotation_z",
        "mu",
        "lamda",
        "cavity_pressure",
    ]
    ro_parameters = ["volume"]

    class EllipsoidalSourceDialog(SourceEditDialog):
        def __init__(self, delegate, *args, **kwargs):
            SourceEditDialog.__init__(
                self, delegate, ui_file="ellipsoid_source.ui", *args, **kwargs
            )

            def setLabel(method, fmt, value, suffix=""):
                method(fmt.format(value) + suffix)

            self.volume.setValue = lambda v: setLabel(
                self.volume.setText, "{:.2e}", v, " m<sup>3</sup>"
            )

            self.getSourceParameters()

    ROIWidget = PointSourceROI
    EditDialog = EllipsoidalSourceDialog

    @staticmethod
    def getRepresentedSource(sandbox):
        return EllipsoidSource()

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>Dimensions</td><td>
        x: {source.length_x:.0f} m;
        y: {source.length_y:.0f} m;
        z: {source.length_z:.0f} m
    </td>
</tr><tr>
    <td>Rotations</td><td>
        x: {source.rotation_x:.1f}&deg;;
        y: {source.rotation_y:.1f}&deg;;
        z: {source.rotation_z:.1f}&deg;
    </td>
</tr><tr style="font-weight: bold;">
    <td>Volume:</td><td>{source.volume:.2e} m<sup>3</sup></td>
</tr></table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)


class PointCompoundSourceDelegate(SourceDelegate):

    __represents__ = "PointCompoundSource"

    display_backend = "Compound Model"
    display_name = "PointCompoundSource"

    parameters = [
        "easting",
        "northing",
        "depth",
        "dVx",
        "dVy",
        "dVz",
        "rotation_x",
        "rotation_y",
        "rotation_z",
        "nu",
    ]
    ro_parameters = ["volume"]

    class CompoundSourceDialog(SourceEditDialog):
        def __init__(self, delegate, *args, **kwargs):
            SourceEditDialog.__init__(
                self, delegate, ui_file="pCDM_source.ui", *args, **kwargs
            )

            def setLabel(method, fmt, value, suffix=""):
                method(fmt.format(value) + suffix)

            self.volume.setValue = lambda v: setLabel(
                self.volume.setText, "{:.2e}", v, " m<sup>3</sup>"
            )

            self.getSourceParameters()

    ROIWidget = PointSourceROI
    EditDialog = CompoundSourceDialog

    @staticmethod
    def getRepresentedSource(sandbox):
        return PointCompoundSource()

    def formatListItem(self):
        item = """
<span style="font-weight: bold; font-style: oblique">
    {idx}. {delegate.display_name}
    <span style="color: #616161;">
        ({delegate.display_backend})
    </span>
</span>
<table style="color: #616161; font-size: small;">
<tr>
    <td>Depth:</td><td>{source.depth:.2f} m</td>
</tr><tr>
    <td>Volume Change</td><td>
        x: {source.dVx:.0f} m<sup>3</sup>;
        y: {source.dVy:.0f} m<sup>3</sup>;
        z: {source.dVz:.0f} m<sup>3</sup>
    </td>
</tr><tr>
    <td>Rotations</td><td>
        x: {source.rotation_x:.1f}&deg;;
        y: {source.rotation_y:.1f}&deg;;
        z: {source.rotation_z:.1f}&deg;
    </td>
</tr><tr style="font-weight: bold;">
    <td>Volume:</td><td>{source.volume:.2e} m<sup>3</sup></td>
</tr></table>
"""
        return item.format(idx=self.index.row() + 1, delegate=self, source=self.source)



import numpy as np
import pyqtgraph as pg
from PyQt5 import QtCore, QtWidgets

from kite.qt_utils import loadUi

from ..util import get_resource

d2r = np.pi / 180.0
r2d = 180.0 / np.pi


class RectangularSourceROI(pg.ROI):

    newSourceParameters = QtCore.pyqtSignal(object)

    pen_outline = pg.mkPen((46, 125, 50, 100), width=1.25)
    pen_handle = pg.mkPen(((38, 50, 56, 100)), width=1.25)
    pen_highlight = pg.mkPen((52, 175, 60), width=2.5)

    def __init__(self, delegate):
        self.delegate = delegate
        self.source = self.delegate.source
        source = self.source

        pg.ROI.__init__(
            self,
            pos=pg.Point(source.outline()[1]),
            size=pg.Point(source.width, source.length),
            angle=-source.strike,
            invertible=False,
            pen=self.pen_outline,
        )
        self.handlePen = self.pen_handle

        self.addScaleRotateHandle([0, 0], [0, 1])
        self.addScaleRotateHandle([0, 1], [0, 0])
        self.addScaleHandle([1, 0.5], [0, 0.5], lockAspect=False)

        for h in self.handles:
            h["item"].sigClicked.connect(self.sigRegionChangeStarted.emit)

        self.delegate.sourceParametersChanged.connect(self.updateROIPosition)
        self.sigRegionChangeFinished.connect(self.setSourceParametersFromROI)

        self.setAcceptedMouseButtons(QtCore.Qt.RightButton)
        self.sigClicked.connect(self.showEditingDialog)

    @QtCore.pyqtSlot()
    def setSourceParametersFromROI(self):

        strike = float((-self.angle()) % 360)
        width = float(self.size().x())
        length = float(self.size().y())

        northing = float(self.pos().y() + np.cos(strike * d2r) * length / 2)
        easting = float(self.pos().x() + np.sin(strike * d2r) * length / 2)

        north_shift, east_shift = self.source.getSandboxOffset()
        easting -= east_shift
        northing -= north_shift

        self.newSourceParameters.emit(
            dict(
                strike=strike,
                width=width,
                length=length,
                easting=easting,
                northing=northing,
            )
        )

    @QtCore.pyqtSlot()
    def updateROIPosition(self):
        source = self.source
        width = source.width

        self.setPos(pg.Point(source.outline()[1]), finish=False)
        self.setSize(pg.Point(width, source.length), finish=False)
        self.setAngle(-source.strike, finish=False)

    @QtCore.pyqtSlot(bool)
    def highlightROI(self, highlight):
        self.setMouseHover(highlight)

    @QtCore.pyqtSlot()
    def showEditingDialog(self, *args):
        self.delegate.getEditingDialog().show()

    def _makePen(self):
        # Generate the pen color for this ROI based on its current state.
        if self.mouseHovering:
            return self.pen_highlight
        else:
            return self.pen


class PointSourceROI(pg.EllipseROI):

    newSourceParameters = QtCore.pyqtSignal(object)

    pen_outline = pg.mkPen((46, 125, 50, 100), width=1.25)
    pen_handle = pg.mkPen(((38, 50, 56, 100)), width=1.25)
    pen_highlight = pg.mkPen((52, 175, 60), width=2.5)

    def __init__(self, delegate):
        self.delegate = delegate
        self.source = self.delegate.source
        source = self.source

        size = 3000.0
        super().__init__(
            pos=(source.easting - size / 2, source.northing - size / 2),
            size=size,
            invertible=False,
            pen=self.pen_outline,
        )
        self.handlePen = self.pen_handle
        self.aspectLocked = True

        for h in self.handles:
            hi = h["item"]
            hi.disconnectROI(self)

        self.handles = []
        # self.setFlag(self.ItemIgnoresTransformations)
        self.setScale(1.0)

        self.delegate.sourceParametersChanged.connect(self.updateROIPosition)
        self.sigRegionChangeFinished.connect(self.setSourceParametersFromROI)

        self.setAcceptedMouseButtons(QtCore.Qt.RightButton)
        self.sigClicked.connect(self.showEditingDialog)

    @QtCore.pyqtSlot()
    def setSourceParametersFromROI(self):
        north_shift, east_shift = self.source.getSandboxOffset()
        self.newSourceParameters.emit(
            dict(
                easting=float(self.pos().x() + self.size().x() / 2 - east_shift),
                northing=float(self.pos().y() + self.size().y() / 2 - north_shift),
            )
        )

    @QtCore.pyqtSlot()
    def updateROIPosition(self):
        source = self.source
        self.setPos(
            pg.Point(
                (
                    source.easting - self.size().x() / 2,
                    source.northing - self.size().x() / 2,
                )
            ),
            finish=False,
        )

    @QtCore.pyqtSlot(bool)
    def highlightROI(self, highlight):
        self.setMouseHover(highlight)

    @QtCore.pyqtSlot()
    def showEditingDialog(self, *args):
        self.delegate.getEditingDialog().show()

    def _makePen(self):
        # Generate the pen color for this ROI based on its current state.
        if self.mouseHovering:
            return self.pen_highlight
        else:
            return self.pen


class SourceDelegate(QtCore.QObject):

    __represents__ = "SourceToImplement"

    sourceParametersChanged = QtCore.pyqtSignal()
    highlightROI = QtCore.pyqtSignal(bool)

    parameters = ["List", "of", "parameters", "from", "kite.source"]
    ro_parameters = ["Read-Only", "parameters"]

    ROIWidget = None  # For use in pyqtgraph
    EditDialog = None  # QDialog to edit the source

    def __init__(self, model, source, index, parent=None):
        super().__init__(parent)
        self.source = source
        self.model = model
        self.index = index
        self.rois = []

        self.editing_dialog = None

        if model.selection_model is not None:
            self.setSelectionModel()

        self.model.selectionModelChanged.connect(self.setSelectionModel)

    @staticmethod
    def getRepresentedSource(sandbox):
        raise NotImplementedError()

    def getEditingDialog(self):
        if self.editing_dialog is None:
            self.editing_dialog = self.EditDialog(self)
        return self.editing_dialog

    def formatListItem(self):
        raise NotImplementedError()
        return "<b>Richtext</b> Source details"

    def getROIItem(self):
        src = self.ROIWidget(self)

        src.newSourceParameters.connect(self.updateModelParameters)
        src.sigRegionChangeStarted.connect(self.highlightItem)
        src.sigHoverEvent.connect(self.highlightItem)
        self.highlightROI.connect(src.highlightROI)

        self.rois.append(src)
        return src

    @QtCore.pyqtSlot()
    def highlightItem(self):
        if self.model.selection_model is not None:
            self.model.selection_model.setCurrentIndex(
                self.index, QtCore.QItemSelectionModel.SelectCurrent
            )

    @QtCore.pyqtSlot()
    def emitHighlightROI(self):
        selected = self.model.selection_model.currentIndex()
        if (
            self.index.row() == selected.row()
            and self.index.column() == selected.column()
        ):
            self.highlightROI.emit(True)
        else:
            self.highlightROI.emit(False)

    @QtCore.pyqtSlot()
    def setSelectionModel(self):
        self.model.selection_model.currentChanged.connect(self.emitHighlightROI)

    @QtCore.pyqtSlot(object)
    def updateModelParameters(self, parameters):
        self.model.setItemData(self.index, parameters)

    def setSourceParameters(self, parameters):
        for param, value in parameters.items():
            self.source.__setattr__(param, value)
        self.source.parametersUpdated()
        self.sourceParametersChanged.emit()

    @QtCore.pyqtSlot()
    def getSourceParameters(self):
        params = {}
        for param in self.parameters + self.ro_parameters:
            params[param] = self.source.__getattribute__(param)
        return params


class SourceEditDialog(QtWidgets.QDialog):
    def __init__(self, delegate, ui_file, *args, **kwargs):
        QtWidgets.QDialog.__init__(self, *args, **kwargs)
        loadUi(get_resource(ui_file), self)
        self.delegate = delegate

        self.delegate.sourceParametersChanged.connect(self.getSourceParameters)
        self.applyButton.released.connect(self.setSourceParameters)
        self.okButton.released.connect(self.setSourceParameters)
        self.okButton.released.connect(self.close)

    @QtCore.pyqtSlot()
    def getSourceParameters(self):
        for param, value in self.delegate.getSourceParameters().items():
            self.__getattribute__(param).setValue(value)

    @QtCore.pyqtSlot()
    def setSourceParameters(self):
        params = {}
        for param in self.delegate.parameters:
            params[param] = self.__getattribute__(param).value()
        self.delegate.updateModelParameters(params)



from PyQt5 import QtCore, QtGui, QtWidgets

from .sources import __sources__
from .util import SourceEditorDialog

km = 1e3


class SourcesListDock(QtWidgets.QDockWidget):
    def __init__(self, sandbox, *args, **kwargs):
        QtWidgets.QDockWidget.__init__(self, "Sources", *args, **kwargs)
        self.sandbox = sandbox

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(3, 3, 3, 3)
        sources = SourcesList(sandbox, parent=self)
        sources_add_menu = SourcesAddButton(sandbox)

        self.setFeatures(
            QtWidgets.QDockWidget.DockWidgetFloatable
            | QtWidgets.QDockWidget.DockWidgetMovable
        )

        self.widget = QtWidgets.QWidget()
        self.widget.setLayout(layout)
        self.widget.layout().addWidget(sources)
        self.widget.layout().addWidget(sources_add_menu)

        self.setWidget(self.widget)


class SourcesAddButton(QtWidgets.QToolButton):
    class SourcesAddMenu(QtWidgets.QMenu):
        def __init__(self, sandbox, *args, **kwargs):
            QtWidgets.QMenu.__init__(self)
            self.setTitle("Add Source")
            self.sandbox = sandbox

            backends = set(src.display_backend for src in __sources__)

            for ibe, be in enumerate(backends):
                self.addSection(be)

                for src in [s for s in __sources__ if s.display_backend == be]:
                    self.addSourceDelegate(src)

                if ibe is not len(backends) - 1:
                    self.addSeparator()

        def addSourceDelegate(self, src):
            def addSource():
                source = src.getRepresentedSource(self.sandbox)
                source.lat = self.sandbox.frame.llLat
                source.lon = self.sandbox.frame.llLon
                source.easting = self.sandbox.frame.Emeter[-1] / 2
                source.northing = self.sandbox.frame.Nmeter[-1] / 2
                source.depth = 4 * km

                if source:
                    self.sandbox.addSource(source)

            action = QtWidgets.QAction(src.display_name, self)
            action.setToolTip(
                '<span style="font-family: monospace;">'
                "%s</span>" % src.__represents__
            )

            action.triggered.connect(addSource)
            self.addAction(action)
            return action

        def addSection(self, text):
            action = QtWidgets.QAction(text, self)
            # action.setSeparator(True)
            font = action.font()
            font.setPointSize(9)
            font.setItalic(True)
            action.setFont(font)
            action.setEnabled(False)
            self.addAction(action)
            return action

    def __init__(self, sandbox, parent=None):
        QtWidgets.QToolButton.__init__(self, parent)

        menu = self.SourcesAddMenu(sandbox, self, "Availables sources")

        self.setText("Add Source")
        self.setMenu(menu)

        self.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_FileDialogDetailedView)
        )
        self.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)


class SourcesList(QtWidgets.QListView):
    class SourceItemDelegate(QtWidgets.QStyledItemDelegate):
        def paint(self, painter, option, index):
            options = QtWidgets.QStyleOptionViewItem(option)
            self.initStyleOption(options, index)

            style = (
                QtWidgets.QApplication.style()
                if options.widget is None
                else options.widget.style()
            )

            doc = QtGui.QTextDocument()
            doc.setHtml(options.text)

            options.text = ""
            style.drawControl(QtWidgets.QStyle.CE_ItemViewItem, options, painter)

            ctx = QtGui.QAbstractTextDocumentLayout.PaintContext()

            textRect = style.subElementRect(
                QtWidgets.QStyle.SE_ItemViewItemText, options, options.widget
            )
            painter.save()
            painter.translate(textRect.topLeft())
            painter.setClipRect(textRect.translated(-textRect.topLeft()))
            doc.documentLayout().draw(painter, ctx)

            painter.restore()

        def sizeHint(self, option, index):
            options = QtWidgets.QStyleOptionViewItem(option)
            self.initStyleOption(options, index)

            doc = QtGui.QTextDocument()
            doc.setHtml(options.text)
            doc.setTextWidth(options.rect.width())

            return QtCore.QSize(doc.idealWidth(), doc.size().height())

    class SourceContextMenu(QtWidgets.QMenu):
        def __init__(self, parent, idx, *args, **kwargs):
            QtWidgets.QMenu.__init__(self, parent, *args, **kwargs)
            self.parent = parent
            self.sandbox = parent.sandbox
            self.idx = idx

            def removeSource():
                self.sandbox.sources.removeSource(self.idx)

            editAction = self.addAction(
                "Edit", lambda: self.parent.editSource(self.idx)
            )

            self.addMenu(SourcesAddButton.SourcesAddMenu(self.sandbox, self))
            self.addSeparator()

            removeAction = self.addAction(
                self.style().standardIcon(QtWidgets.QStyle.SP_DialogCloseButton),
                "Remove",
                removeSource,
            )

            if self.sandbox.sources.rowCount(QtCore.QModelIndex()) == 0:
                editAction.setEnabled(False)
                removeAction.setEnabled(False)

    def __init__(self, sandbox, *args, **kwargs):
        QtWidgets.QListView.__init__(self, *args, **kwargs)
        self.sandbox = sandbox
        self.setModel(sandbox.sources)
        self.setItemDelegate(self.SourceItemDelegate())
        self.setAlternatingRowColors(True)
        sandbox.sources.setSelectionModel(self.selectionModel())

    def edit(self, idx, trigger, event):
        if (
            trigger == QtWidgets.QAbstractItemView.DoubleClicked
            or trigger == QtWidgets.QAbstractItemView.SelectedClicked
        ):
            self.editSource(idx)
        return False

    @QtCore.pyqtSlot(QtCore.QObject)
    def editSource(self, idx):
        editing_dialog = idx.data(SourceEditorDialog)
        editing_dialog.show()
        editing_dialog.raise_()

    @QtCore.pyqtSlot(object)
    def contextMenuEvent(self, event):
        idx = self.indexAt(event.pos())
        menu = self.SourceContextMenu(self, idx)
        menu.popup(event.globalPos())



import logging
import os
import re

import requests

op = os.path
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("kite.clients")


def _download_file(url, outfile):
    logger.debug("Downloading %s to %s", url, outfile)
    r = requests.get(url)

    with open(outfile, "wb") as f:
        for chunk in r.iter_content(chunk_size=8192):
            if chunk:
                f.write(chunk)
    return outfile


def download_licsar(unw_url, destination="."):
    if not unw_url.endswith(".unw.tif"):
        raise ValueError("%s does not end with .unw.tif!" % unw_url)

    scene_name = op.basename(unw_url)
    url_dir = op.dirname(unw_url)
    product_name = op.basename(unw_url)

    os.makedirs(destination, exist_ok=True)

    logger.info("Downloading surface displacement data from LiCSAR: %s", product_name)
    unw_file = op.join(destination, scene_name)
    _download_file(unw_url, unw_file)

    logger.info("Downloading LOS angles...")
    scene_id = url_dir.split("/")[-3]
    meta_url = op.normpath(op.join(url_dir, "../../metadata"))

    for unit in ("E", "N", "U"):
        fn = "%s.geo.%s.tif" % (scene_id, unit)
        los_url = op.join(meta_url, fn)
        los_url = re.sub(r"^(https:/)\b", r"\1/", los_url, 0)
        outfn = op.normpath(op.join(destination, fn))

        _download_file(los_url, outfn)

    logger.info("Download complete! Open with\n\n\tspool --load=%s", unw_file)


if __name__ == "__main__":
    import sys

    download_licsar(*sys.argv[1:])
